import numpy as np
from os import path
from datetime import datetime
from ..utils import check_supported, convert_str_to_tuple
from .utils import parse_params_values
from .cli_configs import ShowReconstructionTimingsConfig

try:
    import matplotlib.pyplot as plt

    __have_matplotlib__ = True
except ImportError:
    __have_matplotlib__ = False

steps_to_measure = [
    "Reading data",
    "Applying flat-field",
    "Applying double flat-field",
    "Applying CCD corrections",
    "Rotating projections",
    "Performing phase retrieval",
    "Performing unsharp mask",
    "Taking logarithm",
    "Applying radios movements",
    "Normalizing sinograms",
    "Building sinograms",  # deprecated
    "Removing rings on sinograms",
    "Reconstruction",
    "Computing histogram",
    "Saving data",
]


def extract_timings_from_volume_reconstruction_lines(lines, separator=" - "):
    def extract_timestamp(line):
        timestamp = line.split(separator)[0]
        return datetime.strptime(timestamp, "%d-%m-%Y %H:%M:%S")

    def extract_current_step(line):
        return line.split(separator)[-1]

    current_step = extract_current_step(lines[0])
    t1 = extract_timestamp(lines[0])

    res = {}
    for line in lines[1:]:
        line = line.strip()
        if len(line.split(separator)) == 1:
            continue
        timestamp = line.strip().split(separator)[0]
        t2 = datetime.strptime(timestamp, "%d-%m-%Y %H:%M:%S")

        res.setdefault(current_step, [])
        res[current_step].append((t2 - t1).seconds)

        t1 = t2
        current_step = extract_current_step(line)

    return res


def parse_logfile(fname, separator=" - "):
    """
    Returns
    -------
    timings: list of dict
        List of dictionaries: one dict per reconstruction in the log file.
        For each dict, the key is the pipeline step name, and the value is the list of timings for the different chunks.
    """
    with open(fname, "r") as f:
        lines = f.readlines()

    start_text = "Going to reconstruct slices"
    end_text = "Merging reconstructions to"

    start_line = None
    rec_log_bounds = []
    for i, line in enumerate(lines):
        if start_text in line:
            start_line = i
        if end_text in line:
            if start_line is None:
                raise ValueError("Could not find reconstruction start string indicator")
            rec_log_bounds.append((start_line, i))
            rec_file_basename = path.basename(line.split(end_text)[-1])

    results = []
    for bounds in rec_log_bounds:
        start, end = bounds
        timings = {}
        res = extract_timings_from_volume_reconstruction_lines(lines[start:end], separator=separator)
        for step in steps_to_measure:
            if step in res:
                timings[step] = res[step]
        results.append(timings)
    return results


def display_timings_pie(timings, reduce_function=None, cutoffs=None):
    reduce_function = reduce_function or np.median
    cutoffs = cutoffs or (0, np.inf)

    def _format_pie_text(pct, allvals):
        # https://matplotlib.org/stable/gallery/pie_and_polar_charts/pie_and_donut_labels.html
        absolute = int(np.round(pct / 100.0 * np.sum(allvals)))
        return f"{pct:.1f}%\n({absolute:d} s)"

    for run in timings:
        fig = plt.figure()
        pie_labels = []
        pie_sizes = []
        for step_name, step_timings in run.items():
            t = reduce_function(step_timings)
            if t > cutoffs[0] and t < cutoffs[1]:
                # pie_labels.append(step_name)
                pie_labels.append(step_name + "\n(%d s)" % t)
                pie_sizes.append(t)
        ax = fig.subplots()
        # ax.pie(pie_sizes, labels=pie_labels, autopct=lambda pct: _format_pie_text(pct, pie_sizes)) # autopct='%1.1f%%')
        ax.pie(pie_sizes, labels=pie_labels, autopct="%1.1f%%")

        fig.show()
    input("Press any key to continue")


def parse_reclog_cli():
    args = parse_params_values(
        ShowReconstructionTimingsConfig, parser_description="Display reconstruction performances from a log file"
    )
    if not (__have_matplotlib__):
        print("Need matplotlib to use this utility")
        exit(1)

    display_functions = {
        "pie": display_timings_pie,
    }

    logfile = args["logfile"]
    cutoff = args["cutoff"]
    display_type = args["type"]

    check_supported(display_type, display_functions.keys(), "Graphics display type")
    if cutoff is not None:
        cutoff = list(map(float, convert_str_to_tuple(cutoff)))

    timings = parse_logfile(logfile)
    display_functions[display_type](timings, cutoffs=cutoff)

    return 0


if __name__ == "__main__":
    parse_reclog_cli()
    exit(0)



import numpy as np
import pytest
from nabu.reconstruction.filtering import SinoFilter, filter_sinogram
from nabu.cuda.utils import __has_pycuda__
from nabu.opencl.utils import __has_pyopencl__
from nabu.testutils import get_data, generate_tests_scenarios, __do_long_tests__

if __has_pycuda__:
    from nabu.cuda.utils import get_cuda_context
    from nabu.reconstruction.filtering_cuda import CudaSinoFilter
    import pycuda.gpuarray as garray
if __has_pyopencl__:
    import pyopencl.array as parray
    from nabu.opencl.processing import OpenCLProcessing
    from nabu.reconstruction.filtering_opencl import OpenCLSinoFilter, __has_vkfft__

filters_to_test = ["ramlak", "shepp-logan", "tukey"]
padding_modes_to_test = ["constant", "edge"]
if __do_long_tests__:
    filters_to_test = ["ramlak", "shepp-logan", "cosine", "hamming", "hann", "tukey", "lanczos"]
    padding_modes_to_test = SinoFilter.available_padding_modes

tests_scenarios = generate_tests_scenarios(
    {
        "filter_name": filters_to_test,
        "padding_mode": padding_modes_to_test,
        "output_provided": [True, False],
        "truncated_sino": [True, False],
    }
)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.sino = get_data("mri_sino500.npz")["data"]
    cls.sino_truncated = np.ascontiguousarray(cls.sino[:, 160:-160])

    if __has_pycuda__:
        cls.ctx_cuda = get_cuda_context(cleanup_at_exit=False)
        cls.sino_cuda = garray.to_gpu(cls.sino)
        cls.sino_truncated_cuda = garray.to_gpu(cls.sino_truncated)
    if __has_pyopencl__:
        cls.cl = OpenCLProcessing(device_type="all")
        cls.sino_cl = parray.to_device(cls.cl.queue, cls.sino)
        cls.sino_truncated_cl = parray.to_device(cls.cl.queue, cls.sino_truncated)

    yield

    if __has_pycuda__:
        cls.ctx_cuda.pop()


@pytest.mark.usefixtures("bootstrap")
class TestSinoFilter:
    @pytest.mark.parametrize("config", tests_scenarios)
    def test_filter(self, config):
        sino = self.sino if not (config["truncated_sino"]) else self.sino_truncated

        sino_filter = SinoFilter(
            sino.shape,
            filter_name=config["filter_name"],
            padding_mode=config["padding_mode"],
        )
        if config["output_provided"]:
            output = np.zeros_like(sino)
        else:
            output = None
        res = sino_filter.filter_sino(sino, output=output)
        if output is not None:
            assert id(res) == id(output), "when providing output, return value must not change"

        ref = filter_sinogram(
            sino, sino_filter.dwidth_padded, filter_name=config["filter_name"], padding_mode=config["padding_mode"]
        )

        assert np.allclose(res, ref, atol=4e-6)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need Cuda + pycuda to use CudaSinoFilter")
    @pytest.mark.parametrize("config", tests_scenarios)
    def test_cuda_filter(self, config):
        sino = self.sino_cuda if not (config["truncated_sino"]) else self.sino_truncated_cuda
        h_sino = self.sino if not (config["truncated_sino"]) else self.sino_truncated

        sino_filter = CudaSinoFilter(
            sino.shape,
            filter_name=config["filter_name"],
            padding_mode=config["padding_mode"],
            cuda_options={"ctx": self.ctx_cuda},
        )
        if config["output_provided"]:
            output = garray.zeros(sino.shape, "f")
        else:
            output = None
        res = sino_filter.filter_sino(sino, output=output)
        if output is not None:
            assert id(res) == id(output), "when providing output, return value must not change"

        ref = filter_sinogram(
            h_sino, sino_filter.dwidth_padded, filter_name=config["filter_name"], padding_mode=config["padding_mode"]
        )

        assert np.allclose(res.get(), ref, atol=6e-5), "test_cuda_filter: something wrong with config=%s" % (
            str(config)
        )

    @pytest.mark.skipif(
        not (__has_pyopencl__ and __has_vkfft__), reason="Need OpenCL + pyopencl + pyvkfft to use OpenCLSinoFilter"
    )
    @pytest.mark.parametrize("config", tests_scenarios)
    def test_opencl_filter(self, config):
        sino = self.sino_cl if not (config["truncated_sino"]) else self.sino_truncated_cl
        h_sino = self.sino if not (config["truncated_sino"]) else self.sino_truncated

        sino_filter = OpenCLSinoFilter(
            sino.shape,
            filter_name=config["filter_name"],
            padding_mode=config["padding_mode"],
            opencl_options={"ctx": self.cl.ctx},
        )
        if config["output_provided"]:
            output = parray.zeros(self.cl.queue, sino.shape, "f")
        else:
            output = None
        res = sino_filter.filter_sino(sino, output=output)
        if output is not None:
            assert id(res) == id(output), "when providing output, return value must not change"

        ref = filter_sinogram(
            h_sino, sino_filter.dwidth_padded, filter_name=config["filter_name"], padding_mode=config["padding_mode"]
        )

        assert np.allclose(res.get(), ref, atol=6e-5), "test_opencl_filter: something wrong with config=%s" % (
            str(config)
        )



import h5py
import numpy
from typing import Union, Optional
from tomoscan.identifier import BaseIdentifier
from nabu.stitching.config import StitchingConfiguration
from tomoscan.volumebase import VolumeBase
from contextlib import AbstractContextManager


class DumperBase:
    """
    Base class to define all the functions that can be used to save a stitching
    """

    def __init__(self, configuration) -> None:
        assert isinstance(configuration, StitchingConfiguration)
        self._configuration = configuration

    @property
    def configuration(self):
        return self._configuration

    @property
    def output_identifier(self) -> BaseIdentifier:
        raise NotImplementedError("Base class")

    def save_stitched_frame(
        self,
        stitched_frame: numpy.ndarray,
        i_frame: int,
        axis: int,
        **kwargs,
    ):
        self.save_frame_to_disk(
            output_dataset=self.output_dataset,
            index=i_frame,
            stitched_frame=stitched_frame,
            axis=axis,
            region_start=0,
            region_end=None,
        )

    @property
    def output_dataset(self) -> Optional[Union[h5py.VirtualLayout, h5py.Dataset, VolumeBase]]:
        return self._output_dataset

    @output_dataset.setter
    def output_dataset(self, dataset: Optional[Union[h5py.VirtualLayout, h5py.Dataset, VolumeBase]]):
        self._output_dataset = dataset

    @staticmethod
    def save_frame_to_disk(
        output_dataset: Union[h5py.Dataset, h5py.VirtualLayout],
        index: int,
        stitched_frame: Union[numpy.ndarray, h5py.VirtualSource],
        axis: int,
        region_start: int,
        region_end: int,
    ):
        if not isinstance(output_dataset, (h5py.VirtualLayout, h5py.Dataset, numpy.ndarray)):
            raise TypeError(
                f"'output_dataset' should be a 'h5py.Dataset' or a 'h5py.VirtualLayout'. Get {type(output_dataset)}"
            )
        if not isinstance(stitched_frame, (h5py.VirtualSource, numpy.ndarray)):
            raise TypeError(
                f"'stitched_frame' should be a 'numpy.ndarray' or a 'h5py.VirtualSource'. Get {type(stitched_frame)}"
            )
        if isinstance(output_dataset, h5py.VirtualLayout) and not isinstance(stitched_frame, h5py.VirtualSource):
            raise TypeError(
                "output_dataset is an instance of h5py.VirtualLayout and stitched_frame not an instance of h5py.VirtualSource"
            )
        if axis == 0:
            if region_end is not None:
                output_dataset[index, region_start:region_end] = stitched_frame
            else:
                output_dataset[index, region_start:] = stitched_frame
        elif axis == 1:
            if region_end is not None:
                output_dataset[region_start:region_end, index, :] = stitched_frame
            else:
                output_dataset[region_start:, index, :] = stitched_frame
        elif axis == 2:
            if region_end is not None:
                output_dataset[region_start:region_end, :, index] = stitched_frame
            else:
                output_dataset[region_start:, :, index] = stitched_frame
        else:
            raise ValueError(f"provided axis ({axis}) is invalid")

    def create_output_dataset(self):
        """
        function called at the beginning of the stitching to prepare output dataset
        """
        raise NotImplementedError



#
# WIP !
#
# pylint: skip-file
import numpy as np
from ..opencl.utils import __has_pyopencl__
from ..utils import get_opencl_srcfile

if __has_pyopencl__:
    import pyopencl as cl
    from ..opencl.processing import OpenCLProcessing
    from ..opencl.kernel import OpenCLKernel
    from pyopencl.tools import dtype_to_ctype as cl_dtype_to_ctype


class OpenCLRoll:
    def __init__(self, dtype, direction=1, offset=None, **processing_kwargs):
        self.processing = OpenCLProcessing(queue=processing_kwargs.get("queue", None))
        self.dtype = np.dtype(dtype)
        compile_options = ["-DDTYPE=%s" % cl_dtype_to_ctype(self.dtype)]
        self.offset = offset or 0
        self.roll_kernel = OpenCLKernel(
            "roll_forward_x",
            None,
            queue=self.processing.queue,
            filename=get_opencl_srcfile("roll.cl"),
            options=compile_options,
        )
        self.shmem = cl.LocalMemory(self.dtype.itemsize)
        self.direction = direction
        if self.direction < 0:
            self.revert_kernel = OpenCLKernel(
                "revert_array_x",
                None,
                queue=self.processing.queue,
                filename=get_opencl_srcfile("roll.cl"),
                options=compile_options,
            )

    def __call__(self, arr):
        ny, nx = arr.shape
        # Launch one big horizontal workgroup
        wg_x = min((nx - self.offset) // 2, self.processing.queue.device.max_work_group_size)
        local_size = (wg_x, 1, 1)
        global_size = [wg_x, ny]
        if self.direction < 0:
            local_size2 = None
            global_size2 = [nx - self.offset, ny]
            self.revert_kernel(
                arr, np.int32(nx), np.int32(ny), np.int32(self.offset), local_size=local_size2, global_size=global_size2
            )
        self.roll_kernel(
            arr,
            np.int32(nx),
            np.int32(ny),
            np.int32(self.offset),
            self.shmem,
            local_size=local_size,
            global_size=global_size,
        )
        if self.direction < 0:
            self.revert_kernel(
                arr, np.int32(nx), np.int32(ny), np.int32(self.offset), local_size=local_size2, global_size=global_size2
            )
        return arr



import numpy as np
from ..cuda.utils import __has_pycuda__
from ..cuda.processing import CudaProcessing
from ..processing.muladd_cuda import CudaMulAdd
from .shift import VerticalShift


class CudaVerticalShift(VerticalShift):
    def __init__(self, radios_shape, shifts, **cuda_options):
        """
        Vertical Shifter, Cuda backend.
        """
        super().__init__(radios_shape, shifts)
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._init_cuda_arrays()

    def _init_cuda_arrays(self):
        interp_infos_arr = np.zeros((len(self.interp_infos), 2), "f")
        self._d_interp_infos = self.cuda_processing.to_device("_d_interp_infos", interp_infos_arr)
        self._d_radio_new = self.cuda_processing.allocate_array("_d_radio_new", self.radios_shape[1:], "f")
        self._d_radio = self.cuda_processing.allocate_array("_d_radio", self.radios_shape[1:], "f")
        self.muladd_kernel = CudaMulAdd(ctx=self.cuda_processing.ctx)

    def apply_vertical_shifts(self, radios, iangles, output=None):
        """
        Parameters
        ----------
        radios: 3D pycuda.gpuarray.GPUArray
            The input radios. If the optional parameter is not given, they are modified in-place
        iangles:  a sequence of integers
            Must have the same lenght as radios.
            It contains the index at which the shift is found in `self.shifts`
            given by `shifts` argument in the initialisation of the object.
        output: 3D pycuda.gpuarray.GPUArray, optional
            If given, it will be modified to contain the shifted radios.
            Must be of the same shape of `radios`.
        """
        self._check(radios, iangles)
        n_a, n_z, n_x = radios.shape
        assert n_z == self.radios_shape[1]
        x_slice = slice(0, n_x)  # slice(None, None)

        def nonempty_subregion(region):
            if region is None:
                return True
            z_slice = region[0]
            return z_slice.stop - z_slice.start > 0

        d_radio_new = self._d_radio_new
        d_radio = self._d_radio

        for ia in iangles:
            d_radio_new.fill(0)
            d_radio[:] = radios[ia, :, :]  # mul-add kernel won't work with pycuda view
            S0, f = self.interp_infos[ia]
            f = np.float32(f)

            s0 = S0
            if s0 > 0:
                # newradio[:-s0] = radio[s0:] * (1 - f)
                dst_region = (slice(0, n_z - s0), x_slice)
                other_region = (slice(s0, n_z), x_slice)
            elif s0 == 0:
                # newradio[:] = radio[s0:] * (1 - f)
                dst_region = None
                other_region = (slice(s0, n_z), x_slice)
            else:
                # newradio[-s0:] = radio[:s0] * (1 - f)
                dst_region = (slice(-s0, n_z), x_slice)
                other_region = (slice(0, n_z + s0), x_slice)

            if all([nonempty_subregion(reg) for reg in [dst_region, other_region]]):
                self.muladd_kernel(
                    d_radio_new,
                    d_radio,
                    1,
                    1 - f,
                    dst_region=dst_region,
                    other_region=other_region,
                )

            s0 = S0 + 1
            if s0 > 0:
                # newradio[:-s0] += radio[s0:] * f
                dst_region = (slice(0, n_z - s0), x_slice)
                other_region = (slice(s0, n_z), x_slice)
            elif s0 == 0:
                # newradio[:] += radio[s0:] * f
                dst_region = None
                other_region = (slice(s0, n_z), x_slice)
            else:
                # newradio[-s0:] += radio[:s0] * f
                dst_region = (slice(-s0, n_z), x_slice)
                other_region = (slice(0, n_z + s0), x_slice)

            if all([nonempty_subregion(reg) for reg in [dst_region, other_region]]):
                self.muladd_kernel(d_radio_new, d_radio, 1, f, dst_region=dst_region, other_region=other_region)

            if output is None:
                radios[ia, :, :] = d_radio_new[:, :]
            else:
                output[ia, :, :] = d_radio_new[:, :]



from ..processing.processing_base import ProcessingBase
from ..utils import MissingComponentError
from .utils import get_opencl_context, __has_pyopencl__

if __has_pyopencl__:
    from .kernel import OpenCLKernel
    import pyopencl as cl
    import pyopencl.array as parray
    from pyopencl.tools import dtype_to_ctype

    OpenCLArray = parray.Array
else:
    OpenCLArray = MissingComponentError("pyopencl")
    dtype_to_ctype = MissingComponentError("pyopencl")


# pylint: disable=E0606
class OpenCLProcessing(ProcessingBase):
    array_class = OpenCLArray
    dtype_to_ctype = dtype_to_ctype

    def __init__(self, ctx=None, device_type="all", queue=None, profile=False, **kwargs):
        """
        Initialie a OpenCLProcessing instance.

        Parameters
        ----------
        ctx: pycuda.driver.Context, optional
            Existing context to use. If provided, do not create a new context.
        cleanup_at_exit: bool, optional
            Whether to clean-up the context at exit.
            Ignored if ctx is not None.
        """
        super().__init__()
        if queue is not None:
            # re-use an existing queue. In this case the this instance is mostly for convenience
            ctx = queue.context
        if ctx is None:
            self.ctx = get_opencl_context(device_type=device_type, **kwargs)
        else:
            self.ctx = ctx
        if queue is None:
            queue_init_kwargs = {}
            if profile:
                queue_init_kwargs = {"properties": cl.command_queue_properties.PROFILING_ENABLE}
            queue = cl.CommandQueue(self.ctx, **queue_init_kwargs)
        self.queue = queue
        dev_types = {
            cl.device_type.CPU: "cpu",
            cl.device_type.GPU: "gpu",
            cl.device_type.ACCELERATOR: "accelerator",
            -1: "unknown",
        }
        self.device_type = dev_types.get(self.ctx.devices[0].type, "unknown")

    # TODO push_context, pop_context ?

    def _allocate_array_mem(self, shape, dtype):
        return parray.zeros(self.queue, shape, dtype)

    def kernel(self, kernel_name, filename=None, src=None, automation_params=None, **build_kwargs):
        return OpenCLKernel(
            kernel_name,
            None,
            queue=self.queue,
            filename=filename,
            src=src,
            automation_params=automation_params,
            **build_kwargs,
        )



import numpy as np
from ..cuda.processing import CudaProcessing
from ..utils import get_cuda_srcfile
from ..processing.padding_cuda import CudaPadding
from ..processing.fft_cuda import get_fft_class
from .filtering import SinoFilter


class CudaSinoFilter(SinoFilter):
    default_extra_options = {**SinoFilter.default_extra_options, **{"fft_backend": "vkfft"}}

    def __init__(
        self,
        sino_shape,
        filter_name=None,
        padding_mode="zeros",
        extra_options=None,
        cuda_options=None,
    ):
        self._cuda_options = cuda_options or {}
        self.cuda = CudaProcessing(**self._cuda_options)
        super().__init__(sino_shape, filter_name=filter_name, padding_mode=padding_mode, extra_options=extra_options)
        self._init_kernels()

    def _init_fft(self):
        fft_cls = get_fft_class(self.extra_options["fft_backend"])
        self.fft = fft_cls(
            self.sino_padded_shape,
            dtype=np.float32,
            axes=(-1,),
        )

    def _allocate_memory(self):
        self.d_filter_f = self.cuda.allocate_array("d_filter_f", (self.sino_f_shape[-1],), dtype=np.complex64)
        self.d_sino_padded = self.cuda.allocate_array("d_sino_padded", self.fft.shape)
        self.d_sino_f = self.cuda.allocate_array("d_sino_f", self.fft.shape_out, dtype=np.complex64)

    def set_filter(self, h_filt, normalize=True):
        super().set_filter(h_filt, normalize=normalize)
        self.d_filter_f[:] = self.filter_f[:]

    def _init_kernels(self):
        # pointwise complex multiplication
        fname = get_cuda_srcfile("ElementOp.cu")
        if self.ndim == 2:
            kernel_name = "inplace_complex_mul_2Dby1D"
            kernel_sig = "PPii"
        else:
            kernel_name = "inplace_complex_mul_3Dby1D"
            kernel_sig = "PPiii"
        self.mult_kernel = self.cuda.kernel(kernel_name, filename=fname, signature=kernel_sig)
        self.kern_args = (self.d_sino_f, self.d_filter_f)
        self.kern_args += self.d_sino_f.shape[::-1]
        # padding
        self.padding_kernel = CudaPadding(
            self.sino_shape,
            ((0, 0), (self.pad_left, self.pad_right)),
            mode=self.padding_mode,
            cuda_options=self._cuda_options,
        )

    def filter_sino(self, sino, output=None):
        """
        Perform the sinogram siltering.

        Parameters
        ----------
        sino: numpy.ndarray or pycuda.gpuarray.GPUArray
            Input sinogram (2D or 3D)
        output: pycuda.gpuarray.GPUArray, optional
            Output array.
        no_output: bool, optional
            If set to True, no copy is be done. The resulting data lies
            in self.d_sino_padded.
        """
        self._check_array(sino)
        if not (isinstance(sino, self.cuda.array_class)):
            sino = self.cuda.set_array("sino", sino)
        elif not (sino.flags.c_contiguous):
            # Transfer the device array into another, c-contiguous, device array
            # We can throw an error as well in this case, but often we so something like fbp(radios[:, i, :])
            sino_tmp = self.cuda.allocate_array("sino_contig", sino.shape)
            sino_tmp.set(sino)
            sino = sino_tmp

        # Padding
        self.padding_kernel(sino, output=self.d_sino_padded)

        # FFT
        self.fft.fft(self.d_sino_padded, output=self.d_sino_f)

        # multiply padded sinogram with filter in the Fourier domain
        self.mult_kernel(*self.kern_args)  # TODO tune block size ?

        # iFFT
        self.fft.ifft(self.d_sino_f, output=self.d_sino_padded)

        # return
        if output is None:
            res = self.cuda.allocate_array("output", self.sino_shape)
        else:
            res = output
        if self.ndim == 2:
            res[:] = self.d_sino_padded[:, self.pad_left : self.pad_left + self.dwidth]
        else:
            res[:] = self.d_sino_padded[:, :, self.pad_left : self.pad_left + self.dwidth]
        return res

    __call__ = filter_sino



from ..utils import deprecated_class
from .config_validators import str2bool
from dataclasses import dataclass
import os

#
# Decorators and callback mechanism
#


def use_options(step_name, step_attr):
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]
            if step_name not in self.processing_steps:
                self.__setattr__(step_attr, None)
                return
            self._steps_name2component[step_name] = step_attr
            self._steps_component2name[step_attr] = step_name
            return func(*args, **kwargs)

        return wrapper

    return decorator


def pipeline_step(step_attr, step_desc):
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]
            if getattr(self, step_attr, None) is None:
                return
            self.logger.info(step_desc)
            res = func(*args, **kwargs)
            step_name = self._steps_component2name[step_attr]
            callbacks = self._callbacks.get(step_name, None)
            if callbacks is not None:
                for callback in callbacks:
                    callback(self)
            if self.datadump_manager is not None and step_name in self.datadump_manager.data_dump:
                self.datadump_manager.dump_data_to_file(
                    step_name, self.radios, crop_margin=not (self._radios_were_cropped)
                )
            return res

        return wrapper

    return decorator


#
# sub-region, shapes, etc
#


def get_subregion(sub_region, ndim=3):
    """
    Return a "normalized" sub-region in the form ((start_z, end_z), (start_y, end_y), (start_x, end_x)).

    Parameters
    ----------
    sub_region: tuple
        A tuple of tuples or tuple of integers.

    Notes
    -----
    The parameter "sub_region" is normally a tuple of tuples of integers.
    However it can be more convenient to use tuple of integers.
    This function will attempt at catching the different cases, but will fail if
    'sub_region' contains heterogeneous types (ex. tuples along with int)
    """
    if sub_region is None:
        res = ((None, None),)
    elif hasattr(sub_region[0], "__iter__"):
        if set(map(len, sub_region)) != set([2]):
            raise ValueError("Expected each tuple to be in the form (start, end)")
        res = sub_region
    else:
        if len(sub_region) % 2:
            raise ValueError("Expected even number of elements")
        starts, ends = sub_region[::2], sub_region[1::2]
        res = tuple([(s, e) for s, e in zip(starts, ends)])
    if len(res) != ndim:
        res += ((None, None),) * (ndim - len(res))
    return res


#
# Writer - moved to pipeline.writer
#

from .writer import WriterManager

WriterConfigurator = deprecated_class("WriterConfigurator moved to nabu.pipeline.writer.WriterManager", do_print=True)(
    WriterManager
)


@dataclass
class EnvSettings:
    """This class centralises the definitions, possibly documentation, and access to environnt variable
    driven settings.
    It is meant to be used in the following way:
          from nabu.utils import nabu_env_settings
          if not nabu_env_settings.skip_tomoscan_checks:
                do something
    """

    skip_tomoscan_checks: bool = False


def _get_nabu_environment_variables():
    nabu_env_settings = EnvSettings()
    nabu_env_settings.skip_tomoscan_checks = str2bool(os.getenv("SKIP_TOMOSCAN_CHECK", "0"))
    return nabu_env_settings


nabu_env_settings = _get_nabu_environment_variables()



import pytest

from nabu.stitching.stitcher.single_axis import PROGRESS_BAR_STITCH_VOL_DESC
from nabu.stitching.utils.post_processing import SlurmStitchingFollowerThread


@pytest.mark.parametrize(
    "test_case",
    {
        "dump configuration: 100%|": None,
        f"stitching : 100%|": None,
        f"{PROGRESS_BAR_STITCH_VOL_DESC}: 42%": 42.0,
        f"{PROGRESS_BAR_STITCH_VOL_DESC}: 56% toto: 23%": 56.0,
        "": None,
        "my%": None,
    }.items(),
)
def test_SlurmStitchingFollowerContext(test_case):
    """Test that the conversion from log lines created by tqdm can be read back"""
    str_to_test, expected_result = test_case
    assert SlurmStitchingFollowerThread.cast_progress_line_from_log(str_to_test) == expected_result



import os
from .config import parse_nabu_config_file
from ..utils import is_writeable
from ..resources.logger import Logger, PrinterLogger
from .config import validate_config
from ..resources.dataset_analyzer import analyze_dataset
from .estimators import DetectorTiltEstimator


class ProcessConfigBase:
    """
    A class for describing the Nabu process configuration.
    """

    # Must be overriden by inheriting class
    default_nabu_config = None
    config_renamed_keys = None

    def __init__(
        self,
        conf_fname=None,
        conf_dict=None,
        dataset_info=None,
        create_logger=False,
    ):
        """
        Initialize a ProcessConfig class.

        Parameters
        ----------
        conf_fname: str
            Path to the nabu configuration file. If provided, the parameters
            `conf_dict` is ignored.
        conf_dict: dict
            A dictionary describing the nabu processing steps.
            If provided, the parameter `conf_fname` is ignored.
        dataset_info: DatasetAnalyzer
            A `DatasetAnalyzer` class instance.
        checks: bool, optional, default is True
            Whether to perform checks on configuration and datasets (recommended !)
        remove_unused_radios: bool, optional, default is True
            Whether to remove unused radios, i.e radios present in the dataset,
            but not explicitly listed in the scan metadata.
        create_logger: str or bool, optional
            Whether to create a Logger object. Default is False, meaning that the logger
            object creation is left to the user.
            If set to True, a Logger object is created, and logs will be written
            to the file "nabu_dataset_name.log".
            If set to a string, a Logger object is created, and the logs will be written
            to the file specified by this string.
        """

        # Step (1a): create 'nabu_config'
        self._parse_configuration(conf_fname, conf_dict)
        self._create_logger(create_logger)

        # Step (1b): create 'dataset_info'
        self._browse_dataset(dataset_info)

        # Step (2)
        self._update_dataset_info_with_user_config()

        # Step (3): estimate tilt, CoR, ...
        self._dataset_estimations()

        # Step (4)
        self._coupled_validation()

        # Step (5)
        self._build_processing_steps()

        # Step (6)
        self._configure_save_steps()
        self._configure_resume()

    def _create_logger(self, create_logger):
        if create_logger is False:
            self.logger = PrinterLogger()
            return
        elif create_logger is True:
            dataset_loc = self.nabu_config["dataset"]["location"]
            dataset_fname_rel = os.path.basename(dataset_loc)
            if os.path.isfile(dataset_loc):
                logger_filename = os.path.join(
                    os.path.abspath(os.getcwd()), os.path.splitext(dataset_fname_rel)[0] + "_nabu.log"
                )
            else:
                logger_filename = os.path.join(os.path.abspath(os.getcwd()), dataset_fname_rel + "_nabu.log")
        elif isinstance(create_logger, str):
            logger_filename = create_logger
        else:
            raise ValueError("Expected bool or str for create_logger")
        if not is_writeable(os.path.dirname(logger_filename)):
            self.logger = PrinterLogger()
            self.logger.error("Cannot create logger file %s: no permission to write therein" % logger_filename)
        else:
            self.logger = Logger("nabu", level=self.nabu_config["pipeline"]["verbosity"], logfile=logger_filename)

    def _parse_configuration(self, conf_fname, conf_dict):
        """
        Parse the user configuration and builds a dictionary.

        Parameters
        ----------
        conf_fname: str
            Path to the .conf file. Mutually exclusive with 'conf_dict'
        conf_dict: dict
            Dictionary with the configuration. Mutually exclusive with 'conf_fname'
        """
        if not ((conf_fname is None) ^ (conf_dict is None)):
            raise ValueError("You must either provide 'conf_fname' or 'conf_dict'")
        if conf_fname is not None:
            if not os.path.isfile(conf_fname):
                raise ValueError("No such file: %s" % conf_fname)
            self.conf_fname = conf_fname
            self.conf_dict = parse_nabu_config_file(conf_fname)
        else:
            self.conf_dict = conf_dict
        if self.default_nabu_config is None or self.config_renamed_keys is None:
            raise ValueError(
                "'default_nabu_config' and 'config_renamed_keys' must be specified by classes inheriting from ProcessConfig"
            )
        self.nabu_config = validate_config(
            self.conf_dict,
            self.default_nabu_config,
            self.config_renamed_keys,
        )

    def _browse_dataset(self, dataset_info):
        """
        Browse a dataset and builds a data structure with the relevant information.
        """
        self.logger.debug("Browsing dataset")
        if dataset_info is not None:
            self.dataset_info = dataset_info
        else:
            extra_options = {
                "exclude_projections": self.nabu_config["dataset"]["exclude_projections"],
                "hdf5_entry": self.nabu_config["dataset"]["hdf5_entry"],
                "nx_version": self.nabu_config["dataset"]["nexus_version"],
            }
            self.dataset_info = analyze_dataset(
                self.nabu_config["dataset"]["location"], extra_options=extra_options, logger=self.logger
            )

    def _update_dataset_info_with_user_config(self):
        """
        Update the 'dataset_info' (DatasetAnalyzer class instance) data structure with options from user configuration.
        """
        raise ValueError("Base class")

    def _get_rotation_axis_position(self):
        self.dataset_info.axis_position = self.nabu_config["reconstruction"]["rotation_axis_position"]

    def _update_rotation_angles(self):
        raise ValueError("Base class")

    def _dataset_estimations(self):
        """
        Perform estimation of several parameters like center of rotation and detector tilt angle.
        """
        self.logger.debug("Doing dataset estimations")
        self._get_tilt()
        self._get_cor()

    def _get_cor(self):
        raise ValueError("Base class")

    def _get_tilt(self):
        tilt = self.nabu_config["preproc"]["tilt_correction"]
        if isinstance(tilt, str):  # auto-tilt...
            self.tilt_estimator = DetectorTiltEstimator(
                self.dataset_info,
                do_flatfield=self.nabu_config["preproc"]["flatfield"],
                logger=self.logger,
                autotilt_options=self.nabu_config["preproc"]["autotilt_options"],
            )
            tilt = self.tilt_estimator.find_tilt(tilt_method=tilt)
        self.dataset_info.detector_tilt = tilt

    def _coupled_validation(self):
        """
        Validate together the dataset information and user configuration.
        Update 'dataset_info' and 'nabu_config'
        """
        raise ValueError("Base class")

    def _build_processing_steps(self):
        """
        Build the processing steps, i.e a tuple (steps, options) where
           - steps is a list of str (list of processing steps names)
           - options is a dict with processing options
        """
        raise ValueError("Base class")

    build_processing_steps = _build_processing_steps  # COMPAT.

    def _configure_save_steps(self):
        raise ValueError("Base class")

    def _configure_resume(self):
        raise ValueError("Base class")



import numpy as np
from numpy.polynomial.polynomial import Polynomial, polyval
from .alignment import medfilt2d, plt
from .cor import CenterOfRotation

try:
    import skimage.transform as skt

    __have_skimage__ = True
except ImportError:
    __have_skimage__ = False


class CameraTilt(CenterOfRotation):
    def compute_angle(
        self,
        img_1: np.ndarray,
        img_2: np.ndarray,
        method="1d-correlation",
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
    ):
        """Find the camera tilt, given two opposite images.

        This method finds the tilt between the camera pixel columns and the
        rotation axis, by performing a 1-dimensional correlation between two
        opposite images.

        The output of this function, allows to compute motor movements for
        aligning the camera tilt.

        Parameters
        ----------
        img_1: numpy.ndarray
            First image
        img_2: numpy.ndarray
            Second image, it needs to have been flipped already (e.g. using numpy.fliplr).
        method: str
            Tilt angle computation method. Default is "1d-correlation" (traditional).
            All options are:
              - "1d-correlation": fastest, but works best for small tilts
              - "fft-polar": slower, but works well on all ranges of tilts
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`

        Raises
        ------
        ValueError
            In case images are not 2-dimensional or have different sizes.

        Returns
        -------
        cor_offset_pix: float
            Estimated center of rotation position from the center of the RoI in pixels.
        tilt_deg: float
            Estimated camera tilt angle in degrees.

        Examples
        --------
        The following code computes the center of rotation position for two
        given images in a tomography scan, where the second image is taken at
        180 degrees from the first.

        >>> radio1 = data[0, :, :]
        ... radio2 = np.fliplr(data[1, :, :])
        ... tilt_calc = CameraTilt()
        ... cor_offset, camera_tilt = tilt_calc.compute_angle(radio1, radio2)

        Or for noisy images:

        >>> cor_offset, camera_tilt = tilt_calc.compute_angle(radio1, radio2, median_filt_shape=(3, 3))
        """
        self._check_img_pair_sizes(img_1, img_2)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        img_shape = img_2.shape
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_1 = self._prepare_image(img_1, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)
        img_2 = self._prepare_image(img_2, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape)

        if method.lower() == "1d-correlation":
            return self._compute_angle_1dcorrelation(
                img_1, img_2, padding_mode, peak_fit_radius=peak_fit_radius, high_pass=high_pass, low_pass=low_pass
            )
        elif method.lower() == "fft-polar":
            if not __have_skimage__:
                raise ValueError(
                    'Camera tilt calculation using "fft-polar" is only available with scikit-image.'
                    " Please install the package to use this option."
                )

            return self._compute_angle_fftpolar(
                img_1, img_2, padding_mode, peak_fit_radius=peak_fit_radius, high_pass=high_pass, low_pass=low_pass
            )
        else:
            raise ValueError('Invalid method: %s. Valid options are: "1d-correlation" | "fft-polar"' % method)

    def _compute_angle_1dcorrelation(
        self, img_1: np.ndarray, img_2: np.ndarray, padding_mode=None, peak_fit_radius=1, high_pass=None, low_pass=None
    ):
        cc = self._compute_correlation_fft(
            img_1, img_2, padding_mode, axes=(-1,), high_pass=high_pass, low_pass=low_pass
        )

        img_shape = img_2.shape
        cc_h_coords = np.fft.fftfreq(img_shape[-1], 1 / img_shape[-1])

        (f_vals, fh) = self.extract_peak_regions_1d(cc, peak_radius=peak_fit_radius, cc_coords=cc_h_coords)
        fitted_shifts_h = self.refine_max_position_1d(f_vals, return_all_coeffs=True)
        fitted_shifts_h += fh[1, :]

        # Computing tilt
        fitted_shifts_h = medfilt2d(fitted_shifts_h, kernel_size=3)

        half_img_size = (img_shape[-2] - 1) / 2
        cc_v_coords = np.linspace(-half_img_size, half_img_size, img_shape[-2])
        coeffs_h = Polynomial.fit(cc_v_coords, fitted_shifts_h, deg=1).convert().coef

        tilt_deg = np.rad2deg(-coeffs_h[1] / 2)
        cor_offset_pix = coeffs_h[0] / 2

        if self.verbose:
            self.logger.info(
                "Fitted center of rotation (pixels):",
                cor_offset_pix,
                "and camera tilt (degrees):",
                tilt_deg,
            )
            f, ax = plt.subplots(1, 1)
            self._add_plot_window(f, ax=ax)
            ax.plot(cc_v_coords, fitted_shifts_h)
            ax.plot(cc_v_coords, polyval(cc_v_coords, coeffs_h), "-C1")
            ax.set_title("Correlation peaks")
            plt.show(block=self.extra_options["blocking_plots"])

        return cor_offset_pix, tilt_deg

    def _compute_angle_fftpolar(
        self, img_1: np.ndarray, img_2: np.ndarray, padding_mode=None, peak_fit_radius=1, high_pass=None, low_pass=None
    ):
        img_shape = img_2.shape

        img_fft_1, img_fft_2, filt, _ = self._transform_to_fft(
            img_1, img_2, padding_mode=padding_mode, axes=(-2, -1), low_pass=low_pass, high_pass=high_pass
        )
        if filt is not None:
            img_fft_1 *= filt
            img_fft_2 *= filt

        # abs removes the translation component
        img_fft_1 = np.abs(np.fft.fftshift(img_fft_1, axes=(-2, -1)))
        img_fft_2 = np.abs(np.fft.fftshift(img_fft_2, axes=(-2, -1)))

        # transform to polar coordinates
        img_fft_1 = skt.warp_polar(img_fft_1, scaling="linear", output_shape=img_shape)
        img_fft_2 = skt.warp_polar(img_fft_2, scaling="linear", output_shape=img_shape)

        # only use half of the fft domain
        img_fft_1 = img_fft_1[..., : img_fft_1.shape[-2] // 2, :]
        img_fft_2 = img_fft_2[..., : img_fft_2.shape[-2] // 2, :]

        tilt_pix = self.find_shift(img_fft_1, img_fft_2, shift_axis=-2, return_relative_to_middle=True)
        tilt_deg = -(360 / img_shape[0]) * tilt_pix

        img_1 = skt.rotate(img_1, tilt_deg)
        img_2 = skt.rotate(img_2, -tilt_deg)

        cor_offset_pix = self.find_shift(
            img_1,
            img_2,
            padding_mode=padding_mode,
            peak_fit_radius=peak_fit_radius,
            high_pass=high_pass,
            low_pass=low_pass,
            return_relative_to_middle=True,
        )

        if self.verbose:
            print(
                "Fitted center of rotation (pixels):",
                cor_offset_pix,
                "and camera tilt (degrees):",
                tilt_deg,
            )

        return cor_offset_pix, tilt_deg



__version__ = "2024.2.1"
__nabu_modules__ = [
    "app",
    "cuda",
    "estimation",
    "io",
    "misc",
    "opencl",
    "pipeline",
    "processing",
    "preproc",
    "reconstruction",
    "resources",
    "thirdparty",
    "stitching",
]
version = __version__



from os.path import dirname
import numpy as np
from pycuda.compiler import SourceModule
from ..utils import updiv, get_cuda_srcfile
from ..cuda.processing import CudaProcessing


class MedianFilter:
    """
    A class for performing median filter on GPU with CUDA
    """

    def __init__(
        self,
        shape,
        footprint=(3, 3),
        mode="reflect",
        threshold=None,
        cuda_options=None,
        abs_diff=False,
    ):
        """Constructor of Cuda Median Filter.

        Parameters
        ----------
        shape: tuple
            Shape of the array, in the format (n_rows, n_columns)
        footprint: tuple
            Size of the median filter, in the format (y, x).
        mode: str
            Boundary handling mode. Available modes are:
               - "reflect": cba|abcd|dcb
               - "nearest": aaa|abcd|ddd
               - "wrap": bcd|abcd|abc
               - "constant": 000|abcd|000

            Default is "reflect".
        threshold: float, optional
            Threshold for the "thresholded median filter".
            A thresholded median filter only replaces a pixel value by the median
            if this pixel value is greater or equal than median + threshold.
        abs_diff: bool, optional
            Whether to perform conditional threshold as abs(value - median)

        Notes
        ------
        Please refer to the documentation of the CudaProcessing class for
        the other parameters.
        """
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        self._set_params(shape, footprint, mode, threshold, abs_diff)
        self.cuda_processing.init_arrays_to_none(["d_input", "d_output"])
        self._init_kernels()

    def _set_params(self, shape, footprint, mode, threshold, abs_diff):
        self.data_ndim = len(shape)
        if self.data_ndim == 2:
            ny, nx = shape
            nz = 1
        elif self.data_ndim == 3:
            nz, ny, nx = shape
        else:
            raise ValueError("Expected 2D or 3D data")
        self.shape = shape
        self.Nx = np.int32(nx)
        self.Ny = np.int32(ny)
        self.Nz = np.int32(nz)
        if len(footprint) != 2:
            raise ValueError("3D median filter is not implemented yet")
        if not ((footprint[0] & 1) and (footprint[1] & 1)):
            raise ValueError("Must have odd-sized footprint")
        self.footprint = footprint
        self._set_boundary_mode(mode)
        self.do_threshold = False
        self.abs_diff = abs_diff
        if threshold is not None:
            self.threshold = np.float32(threshold)
            self.do_threshold = True
        else:
            self.threshold = np.float32(0)

    def _set_boundary_mode(self, mode):
        self.mode = mode
        # Some code duplication from convolution
        self._c_modes_mapping = {
            "periodic": 2,
            "wrap": 2,
            "nearest": 1,
            "replicate": 1,
            "reflect": 0,
            "constant": 3,
        }
        mp = self._c_modes_mapping
        if self.mode.lower() not in mp:
            raise ValueError(
                """
                Mode %s is not available. Available modes are:
                %s
                """
                % (self.mode, str(mp.keys()))
            )
        if self.mode.lower() == "constant":
            raise NotImplementedError("mode='constant' is not implemented yet")
        self._c_conv_mode = mp[self.mode]

    def _init_kernels(self):
        # Compile source module
        compile_options = [
            "-DUSED_CONV_MODE=%d" % self._c_conv_mode,
            "-DMEDFILT_X=%d" % self.footprint[1],
            "-DMEDFILT_Y=%d" % self.footprint[0],
            "-DDO_THRESHOLD=%d" % (int(self.do_threshold) + int(self.abs_diff)),
        ]
        fname = get_cuda_srcfile("medfilt.cu")
        nabu_cuda_dir = dirname(fname)
        include_dirs = [nabu_cuda_dir]
        self.sourcemodule_kwargs = {}
        self.sourcemodule_kwargs["options"] = compile_options
        self.sourcemodule_kwargs["include_dirs"] = include_dirs
        with open(fname) as fid:
            cuda_src = fid.read()
        self._module = SourceModule(cuda_src, **self.sourcemodule_kwargs)
        self.cuda_kernel_2d = self._module.get_function("medfilt2d")
        # Blocks, grid
        self._block_size = {2: (32, 32, 1), 3: (16, 8, 8)}[self.data_ndim]  # TODO tune
        self._n_blocks = tuple([updiv(a, b) for a, b in zip(self.shape[::-1], self._block_size)])

    def medfilt2(self, image, output=None):
        """
        Perform a median filter on an image (or batch of images).

        Parameters
        -----------
        images: numpy.ndarray or pycuda.gpuarray
            2D image or 3D stack of 2D images
        output: numpy.ndarray or pycuda.gpuarray, optional
            Output of filtering. If provided, it must have the same shape
            as the input array.
        """
        self.cuda_processing.set_array("d_input", image)
        if output is not None:
            self.cuda_processing.set_array("d_output", output)
        else:
            self.cuda_processing.allocate_array("d_output", self.shape)
        self.cuda_kernel_2d(
            self.cuda_processing.d_input,
            self.cuda_processing.d_output,
            self.Nx,
            self.Ny,
            self.Nz,
            self.threshold,
            grid=self._n_blocks,
            block=self._block_size,
        )
        self.cuda_processing.recover_arrays_references(["d_input", "d_output"])
        if output is None:
            return self.cuda_processing.d_output.get()
        else:
            return output



import numpy
import pytest
from nabu.stitching.sample_normalization import normalize_frame, SampleSide, Method


def test_normalize_frame():
    """
    test normalize_frame function
    """
    with pytest.raises(TypeError):
        normalize_frame("toto", "left", "median")
    with pytest.raises(TypeError):
        normalize_frame(numpy.linspace(0, 100), "left", "median")

    frame = numpy.ones((10, 40))
    frame[:, 15:25] = numpy.arange(1, 101, step=1).reshape((10, 10))

    numpy.testing.assert_array_equal(
        normalize_frame(
            frame=frame,
            side="left",
            method="mean",
            sample_width=10,
            margin_before_sample=2,
        )[:, 15:25],
        numpy.arange(0, 100, step=1).reshape((10, 10)),
    )

    numpy.testing.assert_array_equal(
        normalize_frame(
            frame=frame,
            side="right",
            method="median",
            sample_width=10,
            margin_before_sample=2,
        )[:, 15:25],
        numpy.arange(0, 100, step=1).reshape((10, 10)),
    )

    assert not numpy.array_equal(
        normalize_frame(
            frame=frame,
            side="right",
            method="mean",
            sample_width=10,
            margin_before_sample=20,
        )[:, 15:25],
        numpy.arange(0, 100, step=1).reshape((10, 10)),
    )



import pytest
from nabu.utils import compare_dicts
from nabu.resources.utils import get_quantities_and_units


class TestUnits:
    expected_results = {
        "distance = 1 m ; pixel_size = 2.0 um": {"distance": 1.0, "pixel_size": 2e-6},
        "distance = 1 m ; pixel_size = 2.6 micrometer": {"distance": 1.0, "pixel_size": 2.6e-6},
        "distance = 10 m ; pixel_size = 2e-6 m": {"distance": 10, "pixel_size": 2e-6},
        "distance = .5 m ; pixel_size = 2.6e-4 centimeter": {"distance": 0.5, "pixel_size": 2.6e-6},
        "distance = 10 cm ; pixel_size = 2.5 micrometer ; energy = 1 ev": {
            "distance": 0.1,
            "pixel_size": 2.5e-6,
            "energy": 1.0e-3,
        },
        "distance = 10 cm ; pixel_size = 9.0e-3 millimeter ; energy = 19 kev": {
            "distance": 0.1,
            "pixel_size": 9e-6,
            "energy": 19.0,
        },
    }
    expected_failures = {
        # typo ("ke" instead of "kev")
        "distance = 10 cm ; energy = 10 ke": ValueError("Cannot convert: ke"),
        # No units
        "distance = 10 ; energy = 10 kev": ValueError("not enough values to unpack (expected 2, got 1)"),
        # Unit not separated by space
        "distance = 10m; energy = 10 kev": ValueError("not enough values to unpack (expected 2, got 1)"),
        # Invalid separator
        "distance = 10 m, energy = 10 kev": ValueError("too many values to unpack (expected 2)"),
    }

    def test_conversion(self):
        for test_str, expected_result in self.expected_results.items():
            res = get_quantities_and_units(test_str)
            err_msg = str(
                "Something wrong with quantities/units extraction from '%s': expected %s, got %s"
                % (test_str, str(expected_result), str(res))
            )
            assert compare_dicts(res, expected_result) is None, err_msg

    def test_valid_input(self):
        for test_str, expected_failure in self.expected_failures.items():
            with pytest.raises(type(expected_failure)) as e_info:
                get_quantities_and_units(test_str)
            assert e_info.value.args[0] == str(expected_failure)



from math import sqrt
import numpy as np

from ..cuda.kernel import CudaKernel
from ..cuda.processing import CudaProcessing
from ..reconstruction.filtering_cuda import CudaSinoFilter
from ..utils import get_cuda_srcfile

try:
    import astra

    __have_astra__ = True
except ImportError:
    __have_astra__ = False


class ConebeamReconstructor:
    """
    A reconstructor for cone-beam geometry using the astra toolbox.
    """

    default_extra_options = {
        "axis_correction": None,
        "clip_outer_circle": False,
        "scale_factor": None,
        "filter_cutoff": 1.0,
        "outer_circle_value": 0.0,
        # "use_astra_fdk": True,
        "use_astra_fdk": False,
    }

    def __init__(
        self,
        sinos_shape,
        source_origin_dist,
        origin_detector_dist,
        angles=None,
        volume_shape=None,
        rot_center=None,
        relative_z_position=None,
        pixel_size=None,
        padding_mode="zeros",
        filter_name=None,
        slice_roi=None,
        cuda_options=None,
        extra_options=None,
    ):
        """
        Initialize a cone beam reconstructor. This reconstructor works on slabs of data,
        meaning that one partial volume is obtained from one stack of sinograms.
        To reconstruct a full volume, the reconstructor must be called on a series of sinograms stacks, with
        an updated "relative_z_position" each time.

        Parameters
        -----------
        sinos_shape: tuple
            Shape of the sinograms stack, in the form (n_sinos, n_angles, prj_width)
        source_origin_dist: float
            Distance, in pixel units, between the beam source (cone apex) and the "origin".
            The origin is defined as the center of the sample
        origin_detector_dist: float
            Distance, in pixel units, between the center of the sample and the detector.
        angles: array, optional
            Rotation angles in radians. If provided, its length should be equal to sinos_shape[1].
        volume_shape: tuple of int, optional
            Shape of the output volume slab, in the form (n_z, n_y, n_x).
            If not provided, the output volume slab shape is (sinos_shape[0], sinos_shape[2], sinos_shape[2]).
        rot_center: float, optional
            Rotation axis position. Default is `(detector_width - 1)/2.0`
        relative_z_position: float, optional
            Position of the central slice of the slab, with respect to the full stack of slices.
            By default it is set to zero, meaning that the current slab is assumed in the middle of the stack
        axis_correction: array, optional
            Array of the same size as the number of projections. Each corresponds to a horizontal displacement.
        pixel_size: float or tuple, optional
            Size of the pixel. Possible options:
              - Nothing is provided (default): in this case, all lengths are normalized with respect to the pixel size,
                i.e 'source_origin_dist' and 'origin_detector_dist' should be expressed in pixels (and 'pixel_size' is set to 1).
              - A scalar number is provided: in this case it is the spacing between two pixels (in each dimension)
              - A tuple is provided: in this case it is the spacing between two pixels in both dimensions,
                vertically then horizontally, i.e (detector_spacing_y, detector_spacing_x)
        scale_factor: float, optional
            Post-reconstruction scale factor.
        padding_mode: str, optional
            How to pad the data before applying FDK. By default this is done by astra with zero-padding.
            If padding_mode is other than "zeros", it will be done by nabu and the padded data is passed to astra
            where no additional padding is done.
            Beware that in its current implementation, this option almost doubles the memory needed.
        slice_roi:
            Whether to reconstruct only a region of interest for each horizontal slice.
            This parameter must be in the form (start_x, end_x, start_y, end_y) with no negative values.
            Note that the current implementation just crops the final reconstructed volume,
            i.e there is no speed or memory benefit.
        use_astra_fdk: bool
            Whether to use the native Astra Toolbox FDK implementation.
            If set to False, the cone-beam pre-weighting and projections padding/filtering is done by nabu.
            Note that this parameter is automatically set to False if padding_mode != "zeros".

        Notes
        ------
        This reconstructor is using the astra toolbox [1]. Therefore the implementation uses Astra's
        reference frame, which is centered on the sample (source and detector move around the sample).
        For more information see Fig. 2 of paper [1].

        To define the cone-beam geometry, two distances are needed:
          - Source-origin distance (hereby d1)
          - Origin-detector distance (hereby d2)

        The magnification at distance d2 is m = 1+d2/d1, so given a detector pixel size p_s, the sample voxel size is p_s/m.

        To make things simpler, this class internally uses a different (but equivalent) geometry:
          - d2 is set to zero, meaning that the detector is (virtually) moved to the center of the sample
          - The detector is "re-scaled" to have a pixel size equal to the voxel size (p_s/m)

        Having the detector in the same plane as the sample center simplifies things when it comes to slab-wise reconstruction:
        defining a volume slab (in terms of z_min, z_max) is equivalent to define the detector bounds, like in parallel geometry.


        References
        -----------
        [1] Aarle, Wim & Palenstijn, Willem & Cant, Jeroen & Janssens, Eline & Bleichrodt,
        Folkert & Dabravolski, Andrei & De Beenhouwer, Jan & Batenburg, Kees & Sijbers, Jan. (2016).
        Fast and flexible X-ray tomography using the ASTRA toolbox.
        Optics Express. 24. 25129-25147. 10.1364/OE.24.025129.
        """
        self._configure_extra_options(extra_options)
        self._init_cuda(cuda_options)
        self._set_sino_shape(sinos_shape)
        self._orig_prog_geom = None
        self._init_geometry(
            source_origin_dist,
            origin_detector_dist,
            pixel_size,
            angles,
            volume_shape,
            rot_center,
            relative_z_position,
            slice_roi,
        )
        self._init_fdk(padding_mode, filter_name)
        self._alg_id = None
        self._vol_id = None
        self._proj_id = None

    def _configure_extra_options(self, extra_options):
        self.extra_options = self.default_extra_options.copy()
        self.extra_options.update(extra_options or {})

    def _init_cuda(self, cuda_options):
        cuda_options = cuda_options or {}
        self.cuda = CudaProcessing(**cuda_options)

    def _set_sino_shape(self, sinos_shape):
        if len(sinos_shape) != 3:
            raise ValueError("Expected a 3D shape")
        self.sinos_shape = sinos_shape
        self.n_sinos, self.n_angles, self.prj_width = sinos_shape

    def _init_fdk(self, padding_mode, filter_name):
        self.padding_mode = padding_mode
        self._use_astra_fdk = bool(self.extra_options.get("use_astra_fdk", True))
        self._use_astra_fdk &= padding_mode in ["zeros", "constant", None, "none"]
        if self._use_astra_fdk:
            return
        self.sino_filter = CudaSinoFilter(
            self.sinos_shape[1:],
            filter_name=filter_name,
            padding_mode=self.padding_mode,
            # TODO (?) configure FFT backend
            extra_options={"cutoff": self.extra_options.get("filter_cutoff", 1.0)},
            cuda_options={"ctx": self.cuda.ctx},
        )
        # In astra, FDK pre-weighting does the "n_a/(pi/2) multiplication"
        # TODO not sure where this "magnification **2" factor comes from ?
        mult_factor = self.n_angles / 3.141592 * 2 / (self.magnification**2)
        self.sino_filter.set_filter(self.sino_filter.filter_f * mult_factor, normalize=False)
        #

    def _set_pixel_size(self, pixel_size):
        if pixel_size is None:
            det_spacing_y = det_spacing_x = 1
        elif np.iterable(pixel_size):
            det_spacing_y, det_spacing_x = pixel_size
        else:
            # assuming scalar
            det_spacing_y = det_spacing_x = pixel_size
        self._det_spacing_y = det_spacing_y
        self._det_spacing_x = det_spacing_x

    def _set_slice_roi(self, slice_roi):
        self.slice_roi = slice_roi
        self._vol_geom_n_x = self.n_x
        self._vol_geom_n_y = self.n_y
        self._crop_data = True
        if slice_roi is None:
            return
        start_x, end_x, start_y, end_y = slice_roi
        if roi_is_centered(self.volume_shape[1:], (slice(start_y, end_y), slice(start_x, end_x))):
            # For FDK, astra can only reconstruct subregion centered around the origin
            self._vol_geom_n_x = self.n_x - start_x * 2
            self._vol_geom_n_y = self.n_y - start_y * 2
        else:
            raise NotImplementedError(
                "Cone-beam geometry supports only slice_roi centered around origin (got slice_roi=%s with n_x=%d, n_y=%d)"
                % (str(slice_roi), self.n_x, self.n_y)
            )

    def _init_geometry(
        self,
        source_origin_dist,
        origin_detector_dist,
        pixel_size,
        angles,
        volume_shape,
        rot_center,
        relative_z_position,
        slice_roi,
    ):
        if angles is None:
            self.angles = np.linspace(0, 2 * np.pi, self.n_angles, endpoint=True)
        else:
            self.angles = angles
        if volume_shape is None:
            volume_shape = (self.sinos_shape[0], self.sinos_shape[2], self.sinos_shape[2])
        self.volume_shape = volume_shape
        self.n_z, self.n_y, self.n_x = self.volume_shape
        self.source_origin_dist = source_origin_dist
        self.origin_detector_dist = origin_detector_dist
        self.magnification = 1 + origin_detector_dist / source_origin_dist
        self._set_slice_roi(slice_roi)
        self.vol_geom = astra.create_vol_geom(self._vol_geom_n_y, self._vol_geom_n_x, self.n_z)
        self.vol_shape = astra.geom_size(self.vol_geom)
        self._cor_shift = 0.0
        self.rot_center = rot_center
        if rot_center is not None:
            self._cor_shift = (self.sinos_shape[-1] - 1) / 2.0 - rot_center
        self._set_pixel_size(pixel_size)
        self._axis_corrections = self.extra_options.get("axis_correction", None)
        self._create_astra_proj_geometry(relative_z_position)

    def _create_astra_proj_geometry(self, relative_z_position):
        # This object has to be re-created each time, because once the modifications below are done,
        # it is no more a "cone" geometry but a "cone_vec" geometry, and cannot be updated subsequently
        # (see astra/functions.py:271)
        self.proj_geom = astra.create_proj_geom(
            "cone",
            self._det_spacing_x,
            self._det_spacing_y,
            self.n_sinos,
            self.prj_width,
            self.angles,
            self.source_origin_dist,
            self.origin_detector_dist,
        )
        self.relative_z_position = relative_z_position or 0.0
        # This will turn the geometry of type "cone" into a geometry of type "cone_vec"
        if self._orig_prog_geom is None:
            self._orig_prog_geom = self.proj_geom
        self.proj_geom = astra.geom_postalignment(self.proj_geom, (self._cor_shift, 0))
        # (src, detector_center, u, v) = (srcX, srcY, srcZ, dX, dY, dZ, uX, uY, uZ, vX, vY, vZ)
        vecs = self.proj_geom["Vectors"]

        # To adapt the center of rotation:
        # dX = cor_shift * cos(theta) - origin_detector_dist * sin(theta)
        # dY = origin_detector_dist * cos(theta) + cor_shift * sin(theta)
        if self._axis_corrections is not None:
            # should we check that dX and dY match the above formulas ?
            cor_shifts = self._cor_shift + self._axis_corrections
            vecs[:, 3] = cor_shifts * np.cos(self.angles) - self.origin_detector_dist * np.sin(self.angles)
            vecs[:, 4] = self.origin_detector_dist * np.cos(self.angles) + cor_shifts * np.sin(self.angles)

        # To adapt the z position:
        # Component 2 of vecs is the z coordinate of the source, component 5 is the z component of the detector position
        # We need to re-create the same inclination of the cone beam, thus we need to keep the inclination of the two z positions.
        # The detector is centered on the rotation axis, thus moving it up or down, just moves it out of the reconstruction volume.
        # We can bring back the detector in the correct volume position, by applying a rigid translation of both the detector and the source.
        # The translation is exactly the amount that brought the detector up or down, but in the opposite direction.
        vecs[:, 2] = -self.relative_z_position

    def _set_output(self, volume):
        if volume is not None:
            expected_shape = self.vol_shape  # if not (self._crop_data) else self._output_cropped_shape
            self.cuda.check_array(volume, expected_shape)
            self.cuda.set_array("output", volume)
        if volume is None:
            self.cuda.allocate_array("output", self.vol_shape)
        d_volume = self.cuda.get_array("output")
        z, y, x = d_volume.shape
        self._vol_link = astra.data3d.GPULink(d_volume.ptr, x, y, z, d_volume.strides[-2])
        self._vol_id = astra.data3d.link("-vol", self.vol_geom, self._vol_link)

    def _set_input(self, sinos):
        self.cuda.check_array(sinos, self.sinos_shape)
        self.cuda.set_array("sinos", sinos)  # self.cuda.sinos is now a GPU array
        # TODO don't create new link/proj_id if ptr is the same ?
        # But it seems Astra modifies the input sinogram while doing FDK, so this might be not relevant
        d_sinos = self.cuda.get_array("sinos")

        # self._proj_data_link = astra.data3d.GPULink(d_sinos.ptr, self.prj_width, self.n_angles, self.n_z, sinos.strides[-2])
        self._proj_data_link = astra.data3d.GPULink(
            d_sinos.ptr, self.prj_width, self.n_angles, self.n_sinos, d_sinos.strides[-2]
        )
        self._proj_id = astra.data3d.link("-sino", self.proj_geom, self._proj_data_link)

    def _preprocess_data(self):
        if self._use_astra_fdk:
            return
        d_sinos = self.cuda.sinos
        fdk_preweighting(
            d_sinos, self._orig_prog_geom, relative_z_position=self.relative_z_position, cor_shift=self._cor_shift
        )
        for i in range(d_sinos.shape[0]):
            self.sino_filter.filter_sino(d_sinos[i], output=d_sinos[i])

    def _update_reconstruction(self):
        if self._use_astra_fdk:
            cfg = astra.astra_dict("FDK_CUDA")
        else:
            cfg = astra.astra_dict("BP3D_CUDA")
        cfg["ReconstructionDataId"] = self._vol_id
        cfg["ProjectionDataId"] = self._proj_id
        if self._alg_id is not None:
            astra.algorithm.delete(self._alg_id)
        self._alg_id = astra.algorithm.create(cfg)

    def reconstruct(self, sinos, output=None, relative_z_position=None):
        """
        sinos: numpy.ndarray or pycuda.gpuarray
            Sinograms, with shape (n_sinograms, n_angles, width)
        output: pycuda.gpuarray, optional
            Output array. If not provided, a new numpy array is returned
        relative_z_position: int, optional
            Position of the central slice of the slab, with respect to the full stack of slices.
            By default it is set to zero, meaning that the current slab is assumed in the middle of the stack
        """
        self._create_astra_proj_geometry(relative_z_position)
        self._set_input(sinos)
        self._set_output(output)
        self._preprocess_data()
        self._update_reconstruction()
        astra.algorithm.run(self._alg_id)
        #
        # NB: Could also be done with
        # from astra.experimental import direct_BP3D
        # projector_id = astra.create_projector("cuda3d", self.proj_geom, self.vol_geom, options=None)
        # direct_BP3D(projector_id, self._vol_link, self._proj_data_link)
        #
        result = self.cuda.get_array("output")
        if output is None:
            result = result.get()
        if self.extra_options.get("scale_factor", None) is not None:
            result *= np.float32(self.extra_options["scale_factor"])  # in-place for pycuda
        self.cuda.recover_arrays_references(["sinos", "output"])
        return result

    def __del__(self):
        if getattr(self, "_alg_id", None) is not None:
            astra.algorithm.delete(self._alg_id)
        if getattr(self, "_vol_id", None) is not None:
            astra.data3d.delete(self._vol_id)
        if getattr(self, "_proj_id", None) is not None:
            astra.data3d.delete(self._proj_id)


def selection_is_centered(size, start, stop):
    """
    Return True if (start, stop) define a selection that is centered on the middle of the array.
    """
    if stop > 0:
        stop -= size
    return stop == -start


def roi_is_centered(shape, slice_):
    """
    Return True if "slice_" define a selection that is centered on the middle of the array.
    """
    return all([selection_is_centered(shp, s.start, s.stop) for shp, s in zip(shape, slice_)])


def fdk_preweighting(d_sinos, proj_geom, relative_z_position=0.0, cor_shift=0.0):

    preweight_kernel = CudaKernel(
        "devFDK_preweight",
        filename=get_cuda_srcfile("cone.cu"),
        signature="Piiifffffiii",
    )

    # n_angles, n_z, n_x = d_sinos.shape
    n_z, n_angles, n_x = d_sinos.shape
    det_origin = sqrt(proj_geom["DistanceOriginDetector"] ** 2 + cor_shift**2)

    block = (32, 16, 1)
    grid = (((n_x + 32 - 1) // 32) * ((n_z + 32 - 1) // 32), (n_angles + 16 - 1) // 16, 1)

    preweight_kernel(
        d_sinos,
        np.uint32(n_x),  # unsigned int projPitch,
        np.uint32(0),  # unsigned int startAngle,
        np.uint32(n_angles),  # unsigned int endAngle,
        np.float32(proj_geom["DistanceOriginSource"]),  # float fSrcOrigin,
        np.float32(det_origin),  # float fDetOrigin,
        np.float32(relative_z_position),  # float fZShift,
        np.float32(proj_geom["DetectorSpacingX"]),  # float fDetUSize,
        np.float32(proj_geom["DetectorSpacingY"]),  # float fDetVSize,
        np.int32(n_angles),  # dims.iProjAngles;
        np.int32(n_x),  # dims.iProjU; // number of detectors in the U direction
        np.int32(n_z),  # dims.iProjV // number of detectors in the V direction
        block=block,
        grid=grid,
    )



import math
import numpy as np
from ...resources.logger import LoggerOrPrint
from ...utils import DictToObj


class SpanStrategy:
    def __init__(
        self,
        z_pix_per_proj,
        x_pix_per_proj,
        detector_shape_vh,
        phase_margin_pix,
        projection_angles_deg,
        require_redundancy=False,
        pixel_size_mm=0.1,
        z_offset_mm=0.0,
        logger=None,
        angular_tolerance_steps=0.0,
    ):
        """
        This class does all the accounting for the reconstructible slices, giving for each one the list of the useful angles,
        of the vertical and horizontal shifts,and more ...

        Parameters
        ----------
        z_pix_per_proj : array of floats
            an array of floats with one entry per projection, in pixel units. The values are the vertical displacements of the detector.
            An decreasing z means that the rotation axis is following the positive direction of the detector  vertical axis, which is pointing toward the ground.
            In the experimental setup, the vertical detector axis is pointing toward the ground. Moreover the values are offsetted so that the
            first value is zero. The offset value, in millimiters is z_offset_mm and it is the vertical position of the sample stage relatively
            to the center of the detector. A negative z_offset_mm means that the sample stage is below the detector for the first projection, and this is almost
            always the case, because the sample is above the sample stage.
            A  z_pix=0 value indicates that  the translation-rotation stage "ground" is exactly at the beam height ( supposed to be near the central line of the detector)
            plus z_offset_mm.
            A positive z_pix means that the  translation   stage has been lowered, compared to the first projection, in order to scan higher parts of the sample.
            ( the sample is above the translation-rotation stage).
        x_pix_per_proj : array of floats
            one entry per projection. The horizontal displacement of the detector respect to  the rotation axis.
            A positive x means that the sample shadow on the detector is moving toward the left of the detector.
             (the detector is going right)
        detector_shape_vh : a tuple of two ints
            the vertical and horizontal dimensions
        phase_margin_pix : int
            the  maximum margin needed for the different kernels (phase, unsharp..) otherwhere in the pipeline
        projection_angles_deg :  array of floats
            per each projection the rotation angle of the sample in degree.
        require_redundancy: bool, optional, defaults to False
            It can be set  to True, when there are dead zones in the detector.
            In this case the minimal required angular span is increased from 360 to  2*360
            in order to enforce  the necessary redundancy, which allows the
            correction of the dead zones. The lines which do not satisfy this requirement are not doable.
        z_offset_mm: float
            the vertical position of the sample stage relatively to the center of the detector. A negative z_offset_mm means that the
            sample stage is below the detector for the first projection, and this is almost
            always the case, because the sample is above the sample stage.
        pixel_size_mm:  float, the pixel size in millimiters
           this value is used to give results in units of " millimeters above the sample stage"
           Althougth internally all is calculated in pixel units, it is useful to incorporate
           such information in the spanstrategy object which will then be able to  setup
           reconstruction informations according to several request formats: be they in
           sequential number of reconstructible slices, or millimeters above the stage.
        angular_tolerance_steps: float, defaults to zero
             the angular tolerance, an angular width expressed in units of an angular step, which is tolerated in
             the criteria for deciding if a slice is reconstructable or not
        logger : a logger, optional
        """
        self.logger = LoggerOrPrint(logger)
        self.require_redundancy = require_redundancy
        self.z_pix_per_proj = z_pix_per_proj
        self.x_pix_per_proj = x_pix_per_proj
        self.detector_shape_vh = detector_shape_vh
        self.total_num_images = len(self.z_pix_per_proj)
        self.phase_margin_pix = phase_margin_pix
        self.pix_size_mm = pixel_size_mm
        self.z_offset_mm = z_offset_mm
        self.angular_tolerance_steps = angular_tolerance_steps

        # internally we use increasing angles, so that in all inequalities, that are used
        # to check the span, only such case has to be contemplated.
        # To do so, if needed, we change the sign of the angles.
        if projection_angles_deg[-1] > projection_angles_deg[0]:
            self.projection_angles_deg_internally = projection_angles_deg
            self.angles_are_inverted = False
        else:
            self.projection_angles_deg_internally = -projection_angles_deg
            self.angles_are_inverted = True
        self.projection_angles_deg = projection_angles_deg

        if (
            len(self.x_pix_per_proj) != self.total_num_images
            or len(self.projection_angles_deg_internally) != self.total_num_images
        ):
            message = f"""
            all the arguments z_pix_per_proj, x_pix_per_proj and projection_angles_deg
            must have the same lenght but their lenght were 
            {len(self.z_pix_per_proj) }, {len(self.x_pix_per_proj) }, {len(self.projection_angles_deg_internally) } respectively """

            raise ValueError(message)

        ## informations to be built are initialised to None here below
        """ For a given slice, the procedure for  obtaining the useful angles,  is based on the "sunshine" image,
        The sunshine image has two dimensions, the  second one is the projection number
        while the first runs over the heights of the slices.
        All the logic is based on this image: when a given pixel of this image  is zero, this corresponds
        to a pair   (height,projection)  for which there is no contribution of that projection to that slice.
        """
        self.sunshine_image = None

        """ This will be an array of integer heights. We use a  one-to-one correspondance beween these integers and slices 
         in the reconstructed volume.

         The value of a given item is the vertical coordinate of an horizontal line in in the dectector
         (or above or below (negative values) ). More precisely the line over which the corresponding slice gets projected 
         at the beginning of the scan ( in other words
         for the first  vertical translation entry of the z_pix_per_proj argument) 
         Illustratively, a view height equal to zero corresponds to 
         a slice which projects on row 0 when translation is given by the first value in self.z_pix_per_proj.
         Negative integers for the heights are possible too, according to the direction of translations, which may bring
         above or below the detector. ( Illustratively
         the values in z_pix_per_proj are alway positive for the simple fact that the sample is above
         the roto-translation stage and the stage must be lowered in height for the beam to hit the sample.
         A scan starts  always from a positive z_pix translation but the z_pix value may either decrease
         or increase. In fact for practical reasons, after having done a scan in one direction it is convenient to scan
         also while coming back after a previous scan)
        """
        self.total_view_heights = None

        """ A list wich will contain for every reconstructable heigth, listed in self.total_view_heights, and in the same order,
        a pair of two integer, the first is the first sequential number of the projection for which the height is projected
        inside the detector, the second integer is the last projection number for which the projection occurs inside the detector.
        """
        self.on_detector_projection_intervals = None

        """ All like self.on_detector_projection_intervals, but considering also the margins (phase, unsharp, etc).
        so that  the height is projected inside the detector but while keeping 
        a safe phase_margin_pix distance from the upper and lower border of the detector. 
        """
        self.within_margin_projection_intervals = None

        """ This will contain projection number i_pro, the integer value given by
            ceil(z_pix_per_proj[i_pro])
        This array, together with the here below array self.fract_complement_to_integer_shift_v
        will be used for cropping  when the data are collected for a given to be reconstructed chunk.
        """
        self.integer_shift_v = np.zeros([self.total_num_images], "i")

        """ The fractional vertical shifts are positive floats  < 1.0 pixels. 
        This is the fractional part which, added to self.integer_shift_v, 
        gives back  z_pix_per_proj.
        Together with integer_shift_v, this array is meant to be used, by other modules,  for cropping  
        when the data are collected for a given to be reconstructed chunk."""
        self.fract_complement_to_integer_shift_v = np.zeros([self.total_num_images], "f")

        self._setup_ephemerides()

        self._setup_sunshine()

    def get_doable_span(self):
        """return an object with two properties:
        view_heights_minmax: containining minimum and maximum doable height ( detector reference at iproj=0)
        z_pix_minmax       : containing minimum and maximum heights above the roto-translational sample stage
        """
        vertical_profile = self.sunshine_image.sum(axis=1)

        doable_indexes = np.arange(len(vertical_profile))[vertical_profile > 0]
        vertical_span = doable_indexes.min(), doable_indexes.max()

        if not (vertical_profile[vertical_span[0] : vertical_span[1] + 1] > 0).all():
            message = """ Something wrong occurred in the span preprocessing.
                          It appears that some intermetiade slices are not doables.
                          The doable span should instead be contiguous. Please signal the problem"""
            raise RuntimeError(message)

        view_heights_minmax = self.total_view_heights[list(vertical_span)]

        hmin, hmax = view_heights_minmax
        d_v, d_h = self.detector_shape_vh
        z_min, z_max = (-self.z_pix_per_proj[0] + (d_v - 1) / 2 - hmax, -self.z_pix_per_proj[0] + (d_v - 1) / 2 - hmin)
        res = {
            "view_heights_minmax": view_heights_minmax,
            "z_pix_minmax": (z_min, z_max),
            "z_mm_minmax": (z_min * self.pix_size_mm, z_max * self.pix_size_mm),
        }
        return DictToObj(res)

    def get_informative_string(self):
        doable_span_v = self.get_doable_span()
        if self.z_pix_per_proj[-1] >= self.z_pix_per_proj[-1]:
            direction = "ascending"
        else:
            direction = "descending"

        s = f"""
        Doable vertical span
        --------------------
             The scan has been performed with an {direction} vertical translation of the rotation axis.
              
             The detector vertical axis is up side down.

             Detector reference system at iproj=0:
                from vertical view height  ...   {doable_span_v.view_heights_minmax[0]}
                up to  (included)     ...   {doable_span_v.view_heights_minmax[1]}

                 The slice that projects to the first line of the first projection 
                 corresponds to vertical heigth = 0

             In voxels, the vertical doable span measures: {doable_span_v.z_pix_minmax[1] - doable_span_v.z_pix_minmax[0]}

             And in millimiters above the stage:
                 from vertical height above stage ( mm units)  ...   {doable_span_v.z_mm_minmax[0] - self.z_offset_mm }
                 up to  (included)                                ...  {doable_span_v.z_mm_minmax[1] - self.z_offset_mm }
        """
        return s

    def get_chunk_info(self, span_v_absolute):
        """

        This method returns an object containing all the informations that are needed to reconstruct
        the corresponding chunk

            angle_index_span: a pair of integers indicating the start and the end of useful angles
                              in the array of all the scan angle self.projection_angles_deg
            span_v:  a pair of two integers indicating the start and end of the span relatively  to the lowest value
                     of array self.total_view_heights
            integer_shift_v: an array, containing for each one of the  useful projections of the span,
                             the integer part of vertical shift to be used in cropping,
            fract_complement_to_integer_shift_v :
                             the fractional remainder for cropping.
            z_pix_per_proj: an array, containing for each to be used projection of the span
                            the vertical shift
            x_pix_per_proj: ....the horizontal shit
            angles_rad    :    an array, for each useful projection of the chunk the angle in radian

        Parameters:
        -----------
          span_v_absolute: tuple of integers
             a pair of two integers
              the first  view height ( referred to the detector y axis at iproj=0)
              the second view height
              with the first height smaller than the second.
        """
        span_v = (span_v_absolute[0] - self.total_view_heights[0], span_v_absolute[1] - self.total_view_heights[0])

        sunshine_subset = self.sunshine_image[span_v[0] : span_v[1]]
        angular_profile = sunshine_subset.sum(axis=0)

        angle_indexes = np.arange(len(self.projection_angles_deg_internally))[angular_profile > 0]

        angle_index_span = angle_indexes.min(), angle_indexes.max() + 1
        if not (np.less(0, angular_profile[angle_index_span[0] : angle_index_span[1]]).all()):
            message = """ Something wrong occurred in the span preprocessing.
                          It appears that some intermediate slices are not doables.
                          The doable span should instead be contiguous. Please signal the problem"""
            raise RuntimeError(message)

        chunk_angles_deg = self.projection_angles_deg[angle_indexes]

        my_slicer = slice(angle_index_span[0], angle_index_span[1])
        values = (
            angle_index_span,
            span_v_absolute,
            self.integer_shift_v[my_slicer],
            self.fract_complement_to_integer_shift_v[my_slicer],
            self.z_pix_per_proj[my_slicer],
            self.x_pix_per_proj[my_slicer],
            np.deg2rad(chunk_angles_deg) * (1 - 2 * int(self.angles_are_inverted)),
        )

        key_names = (
            "angle_index_span",
            "span_v",
            "integer_shift_v",
            "fract_complement_to_integer_shift_v",
            "z_pix_per_proj",
            "x_pix_per_proj",
            "angles_rad",
        )

        return DictToObj(dict(zip(key_names, values)))

    def _setup_ephemerides(self):
        """
        A function which will set :

          * self.integer_shift_v
          * self.fract_complement_to_integer_shift_v
          * self.total_view_heights
          * self.on_detector_projection_intervals
          * self.within_margin_projection_intervals
        """

        for i_pro in range(self.total_num_images):
            trans_v = self.z_pix_per_proj[i_pro]
            self.integer_shift_v[i_pro] = math.ceil(trans_v)
            self.fract_complement_to_integer_shift_v[i_pro] = math.ceil(trans_v) - trans_v

        ## The two following line initialize the view height, then considering the vertical translation
        #  the filed of view will be expanded
        total_view_top = self.detector_shape_vh[0]
        total_view_bottom = 0

        total_view_top = max(total_view_top, int(math.ceil(total_view_top + self.z_pix_per_proj.max())))
        total_view_bottom = min(total_view_bottom, int(math.floor(total_view_bottom + self.z_pix_per_proj.min())))

        self.total_view_heights = np.arange(total_view_bottom, total_view_top + 1)

        ## where possible only data from within safe phase margin will be considered. (within_margin)
        ## if it is enough for 360 or more degree.
        ## If it is not enough we'll complete wih data close to the border, provided that data comes from within detector
        ## This will be
        ## the case for the first and last doable slices.
        self.within_margin_projection_intervals = np.zeros(self.total_view_heights.shape + (2,), "i")
        self.on_detector_projection_intervals = np.zeros(self.total_view_heights.shape + (2,), "i")

        self.on_detector_projection_intervals[:, 1] = 0  # empty intervals
        self.within_margin_projection_intervals[:, 1] = 0

        for i_h, height in enumerate(self.total_view_heights):
            previous_is_inside_detector = False
            previous_is_inside_margin = False

            pos_inside_filtered_v = height - self.integer_shift_v

            is_inside_detector = np.less_equal(0, pos_inside_filtered_v)
            is_inside_detector *= np.less(
                pos_inside_filtered_v, self.detector_shape_vh[0] - np.less(0, self.fract_complement_to_integer_shift_v)
            )

            is_inside_margin = np.less_equal(self.phase_margin_pix, pos_inside_filtered_v)
            is_inside_margin *= np.less(
                pos_inside_filtered_v,
                self.detector_shape_vh[0]
                - self.phase_margin_pix
                - np.less(0, self.fract_complement_to_integer_shift_v),
            )

            tmp = np.arange(self.total_num_images)[is_inside_detector]
            if len(tmp):
                self.on_detector_projection_intervals[i_h, :] = (tmp.min(), tmp.max())

            tmp = np.arange(self.total_num_images)[is_inside_detector]
            if len(tmp):
                self.within_margin_projection_intervals[i_h, :] = (tmp.min(), tmp.max())

    def _setup_sunshine(self):
        """It prepares

          * self.sunshine_image

        an image which for every height, contained in self.total_view_heights, and in the same order,
        contains the list of factors, one per every projection of the total list. Each  factor is a backprojection
        weight. A non-doable height corresponds to a line full of zeros.
        A doable height must correspond to a line having one and only one segment of contiguous non zero elements.
        """

        self.sunshine_image = np.zeros(self.total_view_heights.shape + (self.total_num_images,), "f")

        avg_angular_step_deg = np.diff(self.projection_angles_deg_internally).mean()

        projection_angles_for_interp = np.concatenate(
            [
                [self.projection_angles_deg_internally[0] - avg_angular_step_deg],
                self.projection_angles_deg_internally,
                [self.projection_angles_deg_internally[-1] + avg_angular_step_deg],
            ]
        )

        data_container_for_interp = np.zeros_like(projection_angles_for_interp)

        num_angular_periods = math.ceil(
            (self.projection_angles_deg_internally.max() - self.projection_angles_deg_internally.min()) / 360
        )

        for i_h, height in enumerate(self.total_view_heights):
            first_last_on_dect = self.on_detector_projection_intervals[i_h]
            first_last_within_margin = self.within_margin_projection_intervals[i_h]

            if first_last_on_dect[1] == 0:
                # this line never entered in the fov
                continue

            angle_on_dect_first_last = self.projection_angles_deg_internally[first_last_on_dect]
            angle_within_margin_first_last = self.projection_angles_deg_internally[first_last_within_margin]

            # a mask which is positive for angular positions for which the height i_h gets projected within the detector
            mask_on_dect = (
                np.less_equal(angle_on_dect_first_last[0], self.projection_angles_deg_internally)
                * np.less_equal(self.projection_angles_deg_internally, angle_on_dect_first_last[1])
            ).astype("f")

            # a mask which is positive for angular positions for which the height i_h gets projected within the margins
            mask_within_margin = (
                np.less_equal(angle_within_margin_first_last[0], self.projection_angles_deg_internally)
                * np.less_equal(self.projection_angles_deg_internally, angle_within_margin_first_last[1])
            ).astype("f")

            ## create a line which collects contributions from redundant angles
            detector_collector = np.zeros(self.projection_angles_deg_internally.shape, "f")
            margin_collector = np.zeros(self.projection_angles_deg_internally.shape, "f")

            # the following loop tracks, for each projection, the total weight available at the projection angle
            # The additional weight is coming from redundant angles.
            # In this sense the sunshine_image implements a first rudimentary reweighting,
            # which could be in principle used in the full pipeline, althought the regridded pipeline
            # implements a, better, reweighting o its own.
            for i_shift in range(-num_angular_periods, num_angular_periods + 1):
                signus = (
                    1 if (self.projection_angles_deg_internally[-1] > self.projection_angles_deg_internally[0]) else -1
                )

                data_container_for_interp[1:-1] = mask_on_dect

                detector_collector = detector_collector + mask_on_dect * np.interp(
                    (self.projection_angles_deg_internally + i_shift * 360) * signus,
                    signus * projection_angles_for_interp,
                    data_container_for_interp,
                    left=0,
                    right=0,
                )

                data_container_for_interp[1:-1] = mask_within_margin

                margin_collector = margin_collector + mask_within_margin * np.interp(
                    (self.projection_angles_deg_internally + i_shift * 360) * signus,
                    signus * projection_angles_for_interp,
                    data_container_for_interp,
                    left=0,
                    right=0,
                )

            detector_shined_angles = self.projection_angles_deg_internally[detector_collector > 0.99]
            margin_shined_angles = self.projection_angles_deg_internally[margin_collector > 0.99]

            if not len(detector_shined_angles) > 1:
                continue

            avg_step_deg = abs(avg_angular_step_deg)

            if len(margin_shined_angles):
                angular_span_safe_margin = (
                    margin_shined_angles.max()
                    - margin_shined_angles.min()
                    + avg_step_deg * (1.01 + self.angular_tolerance_steps)
                )
            else:
                angular_span_safe_margin = 0

            angular_span_bare_border = (
                detector_shined_angles.max()
                - detector_shined_angles.min()
                + avg_step_deg * (1.01 + self.angular_tolerance_steps)
            )

            if not self.require_redundancy:
                if angular_span_safe_margin >= 360:
                    self.sunshine_image[i_h] = margin_collector
                elif angular_span_bare_border >= 360:
                    self.sunshine_image[i_h] = detector_collector
            else:
                redundancy_angle_deg = 360
                if angular_span_safe_margin >= 360 and angular_span_safe_margin > 2 * (
                    redundancy_angle_deg + avg_step_deg
                ):
                    self.sunshine_image[i_h] = margin_collector
                elif angular_span_bare_border >= 360 and angular_span_bare_border > 2 * (
                    redundancy_angle_deg + avg_step_deg
                ):
                    self.sunshine_image[i_h] = detector_collector

        sunshine_mask = np.less(0.99, self.sunshine_image)

        self.sunshine_image[np.array([True]) ^ sunshine_mask] = 1.0
        self.sunshine_image[:] = 1 / self.sunshine_image
        self.sunshine_image[np.array([True]) ^ sunshine_mask] = 0.0

        shp = self.sunshine_image.shape
        X, Y = np.meshgrid(np.arange(shp[1]), np.arange(shp[0]))

        condition = self.sunshine_image > 0
        self.sunshine_starts = X.min(axis=1, initial=shp[1], where=condition)
        self.sunshine_ends = X.max(axis=1, initial=0, where=condition)
        self.sunshine_ends[self.sunshine_ends > 0] += 1



import os
import pytest
from nabu.app.reduce_dark_flat import reduce_dark_flat

#####
try:
    from tomoscan.tests.utils import NXtomoMockContext
except ImportError:
    from tomoscan.test.utils import NXtomoMockContext


@pytest.fixture(scope="function")
def hdf5_scan(tmp_path):
    """simple fixture to create a scan and provide it to another function"""
    test_dir = tmp_path / "my_hdf5_scan"
    with NXtomoMockContext(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        yield scan


######


@pytest.mark.parametrize("dark_method", (None, "first", "mean"))
@pytest.mark.parametrize("flat_method", (None, "last", "median"))
def test_reduce_dark_flat_hdf5(tmp_path, hdf5_scan, dark_method, flat_method):  # noqa F811
    """simply test output - processing is tested at tomoscan side"""
    # test with default url
    default_darks_path = os.path.join(hdf5_scan.path, hdf5_scan.get_dataset_basename() + "_darks.hdf5")
    default_flats_path = os.path.join(hdf5_scan.path, hdf5_scan.get_dataset_basename() + "_flats.hdf5")

    assert not os.path.exists(default_darks_path)
    assert not os.path.exists(default_flats_path)
    reduce_dark_flat(
        scan=hdf5_scan,
        dark_method=dark_method,
        flat_method=flat_method,
    )

    if dark_method is not None:
        assert os.path.exists(default_darks_path)
    else:
        assert not os.path.exists(default_darks_path)
    if flat_method is not None:
        assert os.path.exists(default_flats_path)
    else:
        assert not os.path.exists(default_flats_path)

    # make sure if already exists and no overwrite fails
    if dark_method is not None or flat_method is not None:
        with pytest.raises(KeyError):
            reduce_dark_flat(
                scan=hdf5_scan,
                dark_method=dark_method,
                flat_method=flat_method,
                overwrite=False,
            )

    # test with url provided by the user
    tuned_darks_path = os.path.join(tmp_path, "new_folder", "darks.hdf5")
    tuned_flats_path = os.path.join(tmp_path, "new_folder", "flats.hdf5")
    assert not os.path.exists(tuned_darks_path)
    assert not os.path.exists(tuned_flats_path)
    reduce_dark_flat(
        scan=hdf5_scan,
        dark_method=dark_method,
        flat_method=flat_method,
        output_reduced_darks_file=tuned_darks_path,
        output_reduced_flats_file=tuned_flats_path,
    )
    if dark_method is not None:
        assert os.path.exists(tuned_darks_path)
    else:
        assert not os.path.exists(tuned_darks_path)
    if flat_method is not None:
        assert os.path.exists(tuned_flats_path)
    else:
        assert not os.path.exists(tuned_flats_path)



import h5py
import numpy
from typing import Union
from silx.utils.enum import Enum as _Enum
from tomoscan.volumebase import VolumeBase
from tomoscan.esrf.volume.hdf5volume import HDF5Volume
from nabu.io.utils import DatasetReader


class AlignmentAxis2(_Enum):
    """Specific alignment named to help users orienting themself with specific name"""

    CENTER = "center"
    LEFT = "left"
    RIGTH = "right"


class AlignmentAxis1(_Enum):
    """Specific alignment named to help users orienting themself with specific name"""

    FRONT = "front"
    CENTER = "center"
    BACK = "back"


class _Alignment(_Enum):
    """Internal alignment to be used for 2D alignment"""

    LOWER_BOUNDARY = "lower boundary"
    HIGHER_BOUNDARY = "higher boundary"
    CENTER = "center"

    @classmethod
    def from_value(cls, value):
        # cast the AlignmentAxis1 and AlignmentAxis2 values to fit the generic definition
        if value in ("front", "left", AlignmentAxis1.FRONT, AlignmentAxis2.LEFT):
            return _Alignment.LOWER_BOUNDARY
        elif value in ("back", "right", AlignmentAxis1.BACK, AlignmentAxis2.RIGTH):
            return _Alignment.HIGHER_BOUNDARY
        elif value in (AlignmentAxis1.CENTER, AlignmentAxis2.CENTER):
            return _Alignment.CENTER
        else:
            return super().from_value(value)


def align_frame(
    data: numpy.ndarray, alignment: _Alignment, alignment_axis: int, new_aligned_axis_size: int, pad_mode="constant"
):
    """
    Align 2D array to extend if size along `alignment_axis` to `new_aligned_axis_size`.

    :param numpy.ndarray data: data (frame) to align (2D numpy array)
    :param alignment_axis: axis along which we want to align the frame. Must be in (0, 1)
    :param HAlignment alignment: alignment strategy
    :param int new_width: output data width
    """
    if alignment_axis not in (0, 1):
        raise ValueError(f"alignment_axis should be in (0, 1). Get {alignment_axis}")
    alignment = _Alignment.from_value(alignment)

    aligned_axis_size = data.shape[alignment_axis]

    if aligned_axis_size > new_aligned_axis_size:
        raise ValueError(
            f"data.shape[alignment_axis] ({data.shape[alignment_axis]}) > new_aligned_axis_size ({new_aligned_axis_size}). Unable to crop data"
        )
    elif aligned_axis_size == new_aligned_axis_size:
        return data
    else:
        if alignment is _Alignment.CENTER:
            lower_boundary = (new_aligned_axis_size - aligned_axis_size) // 2
            higher_boundary = (new_aligned_axis_size - aligned_axis_size) - lower_boundary
        elif alignment is _Alignment.LOWER_BOUNDARY:
            lower_boundary = 0
            higher_boundary = new_aligned_axis_size - aligned_axis_size
        elif alignment is _Alignment.HIGHER_BOUNDARY:
            lower_boundary = new_aligned_axis_size - aligned_axis_size
            higher_boundary = 0
        else:
            raise ValueError(f"alignment {alignment.value} is not handled")

        assert lower_boundary >= 0, f"pad size must be positive - lower boundary isn't ({lower_boundary})"
        assert higher_boundary >= 0, f"pad size must be positive - higher boundary isn't ({higher_boundary})"

        if alignment_axis == 1:
            return numpy.pad(
                data,
                pad_width=((0, 0), (lower_boundary, higher_boundary)),
                mode=pad_mode,
            )
        elif alignment_axis == 0:
            return numpy.pad(
                data,
                pad_width=((lower_boundary, higher_boundary), (0, 0)),
                mode=pad_mode,
            )
        else:
            raise ValueError("alignment_axis should be in (0, 1)")


def align_horizontally(data: numpy.ndarray, alignment: AlignmentAxis2, new_width: int, pad_mode="constant"):
    """
    Align data horizontally to make sure new data width will ne `new_width`.

    :param numpy.ndarray data: data to align
    :param HAlignment alignment: alignment strategy
    :param int new_width: output data width
    """
    alignment = AlignmentAxis2.from_value(alignment).value
    return align_frame(
        data=data, alignment=alignment, new_aligned_axis_size=new_width, pad_mode=pad_mode, alignment_axis=1
    )


class PaddedRawData:
    """
    Util class to extend a data when necessary
    Must to aplpy to a volume and to an hdf5dataset - array
    The idea behind is to avoid loading all the data in memory
    """

    def __init__(self, data: Union[numpy.ndarray, h5py.Dataset], axis_1_pad_width: tuple) -> None:
        self._axis_1_pad_width = numpy.array(axis_1_pad_width)
        if not (self._axis_1_pad_width.size == 2 and self._axis_1_pad_width[0] >= 0 and self._axis_1_pad_width[1] >= 0):
            raise ValueError(f"'axis_1_pad_width' expects to positive elements. Get {axis_1_pad_width}")
        self._raw_data = data
        self._raw_data_end = None
        # note: for now we return only frames with zeros for padded frames.
        # in the future we could imagine having a method and miror existing volume or extend the closest frame, or get a mean value...
        self._empty_frame = None
        self._dtype = None
        self._shape = None
        self._raw_data_shape = self.raw_data.shape

    @staticmethod
    def get_empty_frame(shape, dtype):
        return numpy.zeros(
            shape=shape,
            dtype=dtype,
        )

    @property
    def empty_frame(self):
        if self._empty_frame is None:
            self._empty_frame = self.get_empty_frame(
                shape=(self.shape[0], 1, self.shape[2]),
                dtype=self.dtype,
            )
        return self._empty_frame

    @property
    def shape(self):
        if self._shape is None:
            self._shape = tuple(
                (
                    self._raw_data_shape[0],
                    numpy.sum(
                        numpy.array(self._axis_1_pad_width),
                    )
                    + self._raw_data_shape[1],
                    self._raw_data_shape[2],
                )
            )
        return self._shape

    @property
    def raw_data(self):
        return self._raw_data

    @property
    def raw_data_start(self):
        return self._axis_1_pad_width[0]

    @property
    def raw_data_end(self):
        if self._raw_data_end is None:
            self._raw_data_end = self._axis_1_pad_width[0] + self._raw_data_shape[1]
        return self._raw_data_end

    @property
    def dtype(self):
        if self._dtype is None:
            self._dtype = self.raw_data.dtype
        return self._dtype

    def __getitem__(self, args):
        if not isinstance(args, tuple) and len(args) == 3:
            raise ValueError("only handles 3D slicing")
        elif not (args[0] == slice(None, None, None) and args[2] == slice(None, None, None)):
            raise ValueError(
                "slicing only handled along axis 1. First and third tuple item are expected to be empty slice as slice(None, None, None)"
            )
        else:
            if numpy.isscalar(args[1]):
                args = (
                    args[0],
                    slice(args[1], args[1] + 1, 1),
                    args[2],
                )

            start = args[1].start
            if start is None:
                start = 0
            stop = args[1].stop
            if stop is None:
                stop = self.shape[1]
            step = args[1].step
            # some test
            if start < 0 or stop < 0:
                raise ValueError("only positive position are handled")
            if start >= stop:
                raise ValueError("start >= stop")
            if stop > self.shape[1]:
                raise ValueError("stop > self.shape[1]")
            if step not in (1, None):
                raise ValueError("for now PaddedVolume only handles steps of 1")

            first_part_array = None
            if start < self.raw_data_start and (stop - start > 0):
                stop_first_part = min(stop, self.raw_data_start)
                first_part_array = numpy.repeat(self.empty_frame, repeats=stop_first_part - start, axis=1)
                start = stop_first_part

            third_part_array = None
            if stop > self.raw_data_end and (stop - start > 0):
                if stop > self.shape[1]:
                    raise ValueError("requested slice is out of boundaries")
                start_third_part = max(start, self.raw_data_end)
                third_part_array = numpy.repeat(self.empty_frame, repeats=stop - start_third_part, axis=1)
                stop = self.raw_data_end

            if start >= self.raw_data_start and stop >= self.raw_data_start and (stop - start > 0):
                second_part_array = self.raw_data[:, start - self.raw_data_start : stop - self.raw_data_start, :]
            else:
                second_part_array = None

            parts = tuple(filter(lambda a: a is not None, (first_part_array, second_part_array, third_part_array)))
            return numpy.hstack(
                parts,
            )



import os
import sys
from multiprocessing import Pool
import numpy as np
from scipy.ndimage import gaussian_filter
import h5py
from silx.io.dictdump import h5todict
from nxtomo.application.nxtomo import NXtomo
from .. import version
from ..utils import get_available_threads, DictToObj
from ..pipeline.estimators import oversample
from .utils import parse_params_values
from .cli_configs import DiagToRotConfig


"""
  The operations here below rely on diag objects which are found in the result of a nab-helical run with the diag_zpro_run set to a number > 0
  They are found in the configuration section of the nabu output, in several sequential dataset, with hdf5 dataset keys 
  which are 0,1,2.... corresponding to all the z-windows ( chunck) for which we have collected contributions at different angles
  which are nothing else but pieces of ready to use preprocessed radio , and this for different angles.
  In other words redundant contributions conccurring at the same prepocessed radiography but comming from different moment 
  of the helical scan are kept separate.
  By forming pairs theta, theta +180
  These Pairs should coincide on an overlapping region after flipping one, 
  where they both have signal ( part of them can be dark if the detector is out of view above or below)
  For each key there is a sequence of radio, the corresponding sequence of weights map, the corresponding z translation, and angles

  The number passed to diag_zpro_run object, is >1, and is interpreted by the diagnostic collection run 
  as the number of wished collected angles between 0 and 180. Lets call it num_0_180

  The collection is not done here. Here we exploit the result of a previous collection to deduce, looking at the correlations, the cor

  An example of collection is this :
    
   |_____ diagnostics
   |               |  
                  |__ 0
                  |    |_ radios       (4*num_0_180, chunky, chunkx) 
                  |    |
                  |    |_ weights      (4*num_0_180, chunky, chunkx) 
                  |    |
                  |    |_ angles       ( 4*num_0_180,)
                       |
                       |_ searched_rad ( 2*num_0_180,)    these are all the searched angles between 0 and 360 in radians
                       |
                       |_ zmm_trans    (  4*num_0_180,)    the z translation in mm
                       |
                       |_ zpix_transl  (  4*num_0_180,)    the z translation in pix
                       |
                       |_ pixes_size_mm  scalar


   Here we follow the evolution of the rotation angle along the scan.
   The final result can be left in its detailed form, giving the found cor at every analysed scan position,
   or the result of the interpolation, giving the cor at the two extremal position of z_translation.

"""


def transform_images(diag, args):
    """
    Filter and transform the radios and the weights.

    Filter the radios, and oversample them along the horizontal line.
    The method in general is similar to the composite cor finding.
    Several overlapping positions are used to match redundant contributions at
    different rotation stages ( theta and theta+180).
    But beforehand it is beneficial to remove low spatial frequencies.
    And we do oversampling on the fly.

    Parameters:
       diag: object
         used member of diag are radios and weights
       args: object
         its member are the application parameters. Here we use only:
             low_pass, high_pass, ovs ( oversampling factor for the horisontal dimension )




    """

    diag.radios[:] = (diag.radios / diag.weights).astype("f")

    new_radios = []

    for ima in diag.radios:
        ima = gaussian_filter(ima, [0, args.low_pass], mode="nearest")
        ima = ima - gaussian_filter(ima, [0, args.high_pass], mode="nearest")
        new_radios.append(ima)

    diag.radios = [oversample(ima, [1, args.ovs]).astype("f") for ima in new_radios]

    diag.weights = [oversample(ima, [1, args.ovs]).astype("f") for ima in diag.weights]


def total_merit_list(arg_tuple):
    """
    builds three lists : all_merits, all_energies, all_z_transl

    For every pair (theta, theta+180 ) add an item to the list which contains:
    for "all_merits" a list of merit, one for every overlap in the overlap_list argument,
    for "all_energies", same logic, but calculating the implied energy, implied in the calculation of the merit,
    for "all_z_transl" we add the averaged z_transl for the considered pair
    Parameters:
       diag: object
         used member of diag are radios, weights and zpix_transl
       args: object
          containing the application parameters.
          Its used members are ovs, high_pass, low_pass
    """

    (diag, overlap_list, args) = arg_tuple

    orig_sy, ovsd_sx = diag.radios[0].shape

    all_merits = []
    all_energies = []
    all_z_transls = []

    # the following two lines are in accordance with the nabu collection layout for diagos
    # there are n_angles_pi in [0,180[, and then again the same number of possibly valid radios
    # (check for nan in z translation) in [180,360[, [360,540[, 540,720[
    n_angles_pi = len(diag.radios) // 4
    n_angles_2pi = len(diag.radios) // 2

    # check for (theta, theta+180 )pairs whose first radio of the pair in in [0,180[ or  [360,540[
    for i in list(range(n_angles_pi)) + list(range(n_angles_2pi, n_angles_2pi + n_angles_pi)):
        merits = []
        energies = []
        z_transl = []

        if (not np.isnan(diag.zpix_transl[i])) and (not np.isnan(diag.zpix_transl[i + n_angles_pi])):
            radio1 = diag.radios[i]
            radio2 = diag.radios[i + n_angles_pi]

            weight1 = diag.weights[i]
            weight2 = diag.weights[i + n_angles_pi]

            for overlap in overlap_list:
                if overlap <= ovsd_sx:
                    my_overlap = overlap
                    my_radio1 = radio1
                    my_radio2 = radio2
                    my_weight1 = weight1
                    my_weight2 = weight2
                else:
                    my_overlap = ovsd_sx - (overlap - ovsd_sx)
                    my_radio1 = np.fliplr(radio1)
                    my_radio2 = np.fliplr(radio2)
                    my_weight1 = np.fliplr(weight1)
                    my_weight2 = np.fliplr(weight2)

                radio_common_left = np.fliplr(my_radio1[:, ovsd_sx - my_overlap :])[
                    :, : -(args.ovs * args.high_pass * 2)
                ]
                radio_common_right = my_radio2[:, ovsd_sx - my_overlap : -(args.ovs * args.high_pass * 2)]
                diff_common = radio_common_right - radio_common_left

                weight_common_left = np.fliplr(my_weight1[:, ovsd_sx - my_overlap :])[
                    :, : -(args.ovs * args.high_pass * 2)
                ]
                weight_common_right = my_weight2[:, ovsd_sx - my_overlap : -(args.ovs * args.high_pass * 2)]
                weight_common = weight_common_right * weight_common_left

                if args.use_l1_norm:
                    merits.append(abs(diff_common * weight_common).astype("d").sum())
                    energies.append(abs(weight_common).astype("d").sum())
                else:
                    merits.append((diff_common * diff_common * weight_common).astype("d").sum())
                    energies.append(
                        (
                            (radio_common_left * radio_common_left + radio_common_right * radio_common_right)
                            * weight_common
                        )
                        .astype("d")
                        .sum()
                    )
        else:
            merits = [0] * (len(overlap_list))
            energies = [0] * (len(overlap_list))

        z_transl = 0.5 * (diag.zpix_transl[i] + diag.zpix_transl[i + n_angles_pi])
        all_z_transls.append(z_transl)

        all_merits.append(merits)
        all_energies.append(energies)

    return all_merits, all_energies, all_z_transls


def find_best_interpolating_line(args):
    (all_z_transl, index_overlap_list_a, index_overlap_list_b, all_energies, all_res) = args

    z_a = np.nanmin(all_z_transl)
    z_b = np.nanmax(all_z_transl)

    best_error = np.nan
    best_off_pair = None

    for index_ovlp_a in index_overlap_list_a:
        for index_ovlp_b in index_overlap_list_b:
            index_ovlps = np.interp(all_z_transl, [z_a, z_b], [index_ovlp_a, index_ovlp_b])
            indexes = (np.arange(all_energies.shape[0]))[~np.isnan(index_ovlps)].astype("i")

            index_ovlps = index_ovlps[~np.isnan(index_ovlps)]
            index_ovlps = np.round_(index_ovlps).astype("i")

            diff_enes = all_res[(indexes, index_ovlps)]
            orig_enes = all_energies[(indexes, index_ovlps)]

            error = (diff_enes / (orig_enes + 1.0e-30)).astype("d").sum()

            if not (error > best_error):
                best_error = error
                best_error_pair = index_ovlp_a, index_ovlp_b

    return best_error, best_error_pair  # pylint: disable=E0606


def main(user_args=None):
    """Find the cor as a function f z translation and write an hdf5 which contains interpolable tables.
    This file can be used subsequently with the correct-rot utility.
    """

    if user_args is None:
        user_args = sys.argv[1:]

    args = DictToObj(
        parse_params_values(
            DiagToRotConfig,
            parser_description=main.__doc__,
            program_version="nabu " + version,
            user_args=user_args,
        )
    )

    if args.near is None:
        if args.original_scan is None:
            raise ValueError(
                "the parameter near was not provided but the original_scan parameter was not provided either"
            )
        if args.entry_name is None:
            raise ValueError(
                "the parameter near was not provided but the entry_name parameter  for the original scan was not provided either"
            )

        scan = NXtomo()
        scan.load(file_path=args.original_scan, data_path=args.entry_name)
        args.near = scan.instrument.detector.x_rotation_axis_pixel_position
    else:
        pass

    args.ovs = 4

    diag_url = os.path.join("/", args.entry_name, "reconstruction/configuration/diagnostics")

    diag_keys = []
    with h5py.File(args.diag_file, "r") as f:
        diag_keys = list(f[diag_url].keys())
        diag_keys = [diag_keys[i] for i in np.argsort(list(map(int, diag_keys)))]

    all_merits = []
    all_energies = []
    all_z_transls = []

    arguments_for_multiprocessing = []

    for i_key, my_key in enumerate(diag_keys):
        diag = DictToObj(h5todict(args.diag_file, os.path.join(diag_url, my_key)))

        args.original_shape = diag.radios[0].shape
        args.zpix_mm = diag.pixel_size_mm

        transform_images(diag, args)

        if i_key == 0:
            orig_sy, ovsd_sx = diag.radios[0].shape  # already transformed here, ovsd_sx is expanded
            args.ovsd_sx = ovsd_sx

            overlap_min = max(4, ovsd_sx - 2 * args.ovs * (args.near + args.near_width))
            overlap_max = min(2 * ovsd_sx - 4, ovsd_sx - 2 * args.ovs * (args.near - args.near_width))

            overlap_list = list(range(int(overlap_min), int(overlap_max) + 1))

            if overlap_min > overlap_max:
                message = f""" There is no safe search range in find_cor once the margins corresponding to the high_pass filter are discarded.
                May be the near value (which is the offset respect to the center of the image) is too big, or too negative,
                in short too close to the borders.
                """
                raise ValueError(message)

        arguments_for_multiprocessing.append((diag, overlap_list, args))  # pylint: disable=E0606

    ncpus = get_available_threads()
    with Pool(processes=ncpus) as pool:
        result_list = pool.map(total_merit_list, arguments_for_multiprocessing)

    for merits, energies, z_transls in result_list:
        all_z_transls.extend(z_transls)
        all_merits.extend(merits)
        all_energies.append(energies)

    n_pairings_with_data = 0
    for en, me in zip(all_merits, all_energies):
        if np.any(me):
            n_pairings_with_data += 1

    if args.linear_interpolation:
        if n_pairings_with_data < 2:
            message = f""" The diagnostics collection has probably been run over a too thin section of the scan
            or you scan does not allow to form pairs of theta, theta+360. I only found {n_pairings_with_data} pairings
            and this is not enough to do correlation + interpolation between sections
            """
            raise RuntimeError(message)
    else:
        if n_pairings_with_data < 1:
            message = f""" The diagnostics collection has probably been run over a too thin section of the scan
            or you scan does not allow to form pairs of theta, theta+360. I only found {n_pairings_with_data}
            pairings
            """
            raise RuntimeError(message)

    # all_merits, all_energies, all_z_transls = zip( result_list )

    # merits, energies, z_transls = total_merit_list(diag, overlap_list, args)

    #     all_z_transls.extend(z_transls)
    #     all_merits.extend(merits)
    #     all_energies.append(energies)

    all_merits = np.array(all_merits)
    all_energies = np.array(all_energies)

    all_merits.shape = -1, len(overlap_list)
    all_energies.shape = -1, len(overlap_list)

    if args.linear_interpolation:
        do_linear_interpolation(args, overlap_list, all_merits, all_energies, all_z_transls)
    else:
        do_height_by_height(args, overlap_list, all_merits, all_energies, all_z_transls)

    return 0


def do_height_by_height(args, overlap_list, all_diff, all_energies, all_z_transl):
    # now we find the best cor for each chunk, or nan if no overlap is found
    z_a = np.min(all_z_transl)
    z_b = np.max(all_z_transl)

    grouped_diff = {}
    grouped_energy = {}

    for diff, energy, z in zip(all_diff, all_energies, all_z_transl):
        found = z
        for key in grouped_diff.keys():
            if abs(key - z) < 2.0:  # these are in pixel units
                found = key
                break
        grouped_diff[found] = grouped_diff.get(found, np.zeros([len(overlap_list)], "f")) + diff
        grouped_energy[found] = grouped_energy.get(found, np.zeros([len(overlap_list)], "f")) + energy

    z_list = list(grouped_energy.keys())
    z_list.sort()

    cor_list = []
    for z in z_list:
        diff = grouped_diff[z]
        energy = grouped_energy[z]

        best_error = np.nan
        best_off = None

        if not np.isnan(z):
            for i_ovlp in range(len(overlap_list)):
                error = diff[i_ovlp] / (energy[i_ovlp] + 1.0e-30)

                if not (error > best_error):
                    best_error = error
                    best_off = i_ovlp

            if best_off is not None:
                offset = (args.ovsd_sx - overlap_list[best_off]) / args.ovs / 2
                sy, sx = args.original_shape
                cor_abs = (sx - 1) / 2 + offset
                cor_list.append(cor_abs)
            else:
                cor_list.append(np.nan)
        else:
            # no overlap  was available for that z
            cor_list.append(np.nan)

    with h5py.File(args.cor_file, "w") as f:
        my_mask = ~np.isnan(np.array(cor_list))
        f["cor"] = np.array(cor_list)[my_mask]
        f["z_pix"] = np.array(z_list)[my_mask]
        f["z_m"] = (np.array(z_list)[my_mask]) * args.zpix_mm * 1.0e-3


def do_linear_interpolation(args, overlap_list, all_res, all_energies, all_z_transl):
    # now we consider all the linear regressions of the offset with z_transl

    ncpus = get_available_threads()

    index_overlap_list = np.arange(len(overlap_list)).astype("i")
    arguments_list = [
        (all_z_transl, piece, index_overlap_list, all_energies, all_res)
        for piece in np.array_split(index_overlap_list, ncpus)
    ]

    with Pool(processes=ncpus) as pool:
        result_list = pool.map(find_best_interpolating_line, arguments_list)

    error_list = [tok[0] for tok in result_list]
    best_pos = np.argmin(error_list)
    best_error, best_error_pair = result_list[best_pos]

    # find the interpolated line
    i_ovlp_a, i_ovlp_b = best_error_pair
    offset_a = (args.ovsd_sx - overlap_list[i_ovlp_a]) / args.ovs / 2
    offset_b = (args.ovsd_sx - overlap_list[i_ovlp_b]) / args.ovs / 2

    sy, sx = args.original_shape

    cor_abs_a = (sx - 1) / 2 + offset_a
    cor_abs_b = (sx - 1) / 2 + offset_b

    z_a = np.nanmin(all_z_transl)
    z_b = np.nanmax(all_z_transl)

    with h5py.File(args.cor_file, "w") as f:
        f["cor"] = np.array([cor_abs_a, cor_abs_b])
        f["z_pix"] = np.array([z_a, z_b])
        f["z_m"] = np.array([z_a, z_b]) * args.zpix_mm * 1.0e-3



# pylint: skip-file

"""
This file is a "GPU" (through cupy) implementation of "remove_all_stripe".
The original method is implemented by Nghia Vo in the algotom project: https://github.com/algotom/algotom/blob/master/algotom/prep/removal.py
The implementation using cupy is done by Viktor Nikitin in the tomocupy project: https://github.com/tomography/tomocupy/blame/main/src/tomocupy/remove_stripe.py
then moved to https://github.com/tomography/tomocupy/blob/main/src/tomocupy/processing/remove_stripe.py

For now we can't rely on off-the-shelf tomocupy as it's not packaged in pypi, and compilation is quite tedious.

License follows.
"""

# *************************************************************************** #
#                  Copyright © 2022, UChicago Argonne, LLC                    #
#                           All Rights Reserved                               #
#                         Software Name: Tomocupy                             #
#                     By: Argonne National Laboratory                         #
#                                                                             #
#                           OPEN SOURCE LICENSE                               #
#                                                                             #
# Redistribution and use in source and binary forms, with or without          #
# modification, are permitted provided that the following conditions are met: #
#                                                                             #
# 1. Redistributions of source code must retain the above copyright notice,   #
#    this list of conditions and the following disclaimer.                    #
# 2. Redistributions in binary form must reproduce the above copyright        #
#    notice, this list of conditions and the following disclaimer in the      #
#    documentation and/or other materials provided with the distribution.     #
# 3. Neither the name of the copyright holder nor the names of its            #
#    contributors may be used to endorse or promote products derived          #
#    from this software without specific prior written permission.            #
#                                                                             #
#                                                                             #
# *************************************************************************** #
#                               DISCLAIMER                                    #
#                                                                             #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS         #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT           #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS           #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT    #
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,      #
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED    #
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR      #
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF      #
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING        #
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS          #
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                #
# *************************************************************************** #

try:
    import pycuda.gpuarray as garray
    import cupy as cp
    import pywt
    from cupyx.scipy.ndimage import median_filter
    from cupyx.scipy import signal
    from cupyx.scipy.ndimage import binary_dilation
    from cupyx.scipy.ndimage import uniform_filter1d
    __have_tomocupy_deringer__ = True
except ImportError as err:
    __have_tomocupy_deringer__ = False
    __tomocupy_deringer_import_error__ = err


###### Ring removal with wavelet filtering (adapted for cupy from pytroch_wavelet package https://pytorch-wavelets.readthedocs.io/)################################################################################

def _reflect(x, minx, maxx):
    """Reflect the values in matrix *x* about the scalar values *minx* and
    *maxx*.  Hence a vector *x* containing a long linearly increasing series is
    converted into a waveform which ramps linearly up and down between *minx*
    and *maxx*.  If *x* contains integers and *minx* and *maxx* are (integers +
    0.5), the ramps will have repeated max and min samples.

    .. codeauthor:: Rich Wareham <rjw57@cantab.net>, Aug 2013
    .. codeauthor:: Nick Kingsbury, Cambridge University, January 1999.

    """
    x = cp.asanyarray(x)
    rng = maxx - minx
    rng_by_2 = 2 * rng
    mod = cp.fmod(x - minx, rng_by_2)
    normed_mod = cp.where(mod < 0, mod + rng_by_2, mod)
    out = cp.where(normed_mod >= rng, rng_by_2 - normed_mod, normed_mod) + minx
    return cp.array(out, dtype=x.dtype)


def _mypad(x, pad, value=0):
    """ Function to do numpy like padding on Arrays. Only works for 2-D
    padding.

    Inputs:
        x (array): Array to pad
        pad (tuple): tuple of (left, right, top, bottom) pad sizes        
    """
    # Vertical only
    if pad[0] == 0 and pad[1] == 0:
        m1, m2 = pad[2], pad[3]
        l = x.shape[-2]
        xe = _reflect(cp.arange(-m1, l+m2, dtype='int32'), -0.5, l-0.5)
        return x[:, :, xe]
    # horizontal only
    elif pad[2] == 0 and pad[3] == 0:
        m1, m2 = pad[0], pad[1]
        l = x.shape[-1]
        xe = _reflect(cp.arange(-m1, l+m2, dtype='int32'), -0.5, l-0.5)
        return x[:, :, :, xe]


def _conv2d(x, w, stride, pad, groups=1):
    """ Convolution (equivalent pytorch.conv2d)
    """
    if pad != 0:
        x = cp.pad(x, ((0, 0), (0, 0), (pad, pad), (pad, pad)), 'constant')

    b,  ci, hi, wi = x.shape
    co, _, hk, wk = w.shape
    ho = int(cp.floor(1 + (hi - hk) / stride[0]))
    wo = int(cp.floor(1 + (wi - wk) / stride[1]))
    out = cp.zeros([b, co, ho, wo], dtype='float32')
    x = cp.expand_dims(x, axis=1)
    w = cp.expand_dims(w, axis=0)
    chunk = ci//groups
    chunko = co//groups
    for g in range(groups):
        for ii in range(hk):
            for jj in range(wk):
                x_windows = x[:, :, g*chunk:(g+1)*chunk, ii:ho *
                              stride[0]+ii:stride[0], jj:wo*stride[1]+jj:stride[1]]
                out[:, g*chunko:(g+1)*chunko] += cp.sum(x_windows *
                                                        w[:, g*chunko:(g+1)*chunko, :, ii:ii+1, jj:jj+1], axis=2)
    return out


def _conv_transpose2d(x, w, stride, pad, bias=None, groups=1):
    """ Transposed convolution (equivalent pytorch.conv_transpose2d)
    """
    b,  co, ho, wo = x.shape
    co, ci, hk, wk = w.shape

    hi = (ho-1)*stride[0]+hk
    wi = (wo-1)*stride[1]+wk
    out = cp.zeros([b, ci, hi, wi], dtype='float32')
    chunk = ci//groups
    chunko = co//groups
    for g in range(groups):
        for ii in range(hk):
            for jj in range(wk):
                x_windows = x[:, g*chunko:(g+1)*chunko]
                out[:, g*chunk:(g+1)*chunk, ii:ho*stride[0]+ii:stride[0], jj:wo*stride[1] +
                    jj:stride[1]] += x_windows * w[g*chunko:(g+1)*chunko, :, ii:ii+1, jj:jj+1]
    if pad != 0:
        out = out[:, :, pad[0]:out.shape[2]-pad[0], pad[1]:out.shape[3]-pad[1]]
    return out


def afb1d(x, h0, h1='zero', dim=-1):
    """ 1D analysis filter bank (along one dimension only) of an image

    Parameters
    ----------
    x (array): 4D input with the last two dimensions the spatial input
    h0 (array): 4D input for the lowpass filter. Should have shape (1, 1,
        h, 1) or (1, 1, 1, w)
    h1 (array): 4D input for the highpass filter. Should have shape (1, 1,
        h, 1) or (1, 1, 1, w)
    dim (int) - dimension of filtering. d=2 is for a vertical filter (called
        column filtering but filters across the rows). d=3 is for a
        horizontal filter, (called row filtering but filters across the
        columns).

    Returns
    -------
    lohi: lowpass and highpass subbands concatenated along the channel
        dimension
    """
    C = x.shape[1]
    # Convert the dim to positive
    d = dim % 4
    s = (2, 1) if d == 2 else (1, 2)
    N = x.shape[d]
    L = h0.size
    L2 = L // 2
    shape = [1, 1, 1, 1]
    shape[d] = L
    h = cp.concatenate([h0.reshape(*shape), h1.reshape(*shape)]*C, axis=0)
    # Calculate the pad size
    outsize = pywt.dwt_coeff_len(N, L, mode='symmetric')
    p = 2 * (outsize - 1) - N + L
    pad = (0, 0, p//2, (p+1)//2) if d == 2 else (p//2, (p+1)//2, 0, 0)
    x = _mypad(x, pad=pad)
    lohi = _conv2d(x, h, stride=s, pad=0, groups=C)
    return lohi


def sfb1d(lo, hi, g0, g1='zero', dim=-1):
    """ 1D synthesis filter bank of an image Array
    """

    C = lo.shape[1]
    d = dim % 4
    L = g0.size
    shape = [1, 1, 1, 1]
    shape[d] = L
    N = 2*lo.shape[d]
    s = (2, 1) if d == 2 else (1, 2)
    g0 = cp.concatenate([g0.reshape(*shape)]*C, axis=0)
    g1 = cp.concatenate([g1.reshape(*shape)]*C, axis=0)
    pad = (L-2, 0) if d == 2 else (0, L-2)
    y = _conv_transpose2d(cp.asarray(lo), cp.asarray(g0), stride=s, pad=pad, groups=C) + \
        _conv_transpose2d(cp.asarray(hi), cp.asarray(g1),
                          stride=s, pad=pad, groups=C)
    return y


class DWTForward():
    """ Performs a 2d DWT Forward decomposition of an image

    Args:
        wave (str): Which wavelet to use.                    
        """

    def __init__(self, wave='db1'):
        super().__init__()

        wave = pywt.Wavelet(wave)
        h0_col, h1_col = wave.dec_lo, wave.dec_hi
        h0_row, h1_row = h0_col, h1_col

        self.h0_col = cp.array(h0_col).astype('float32')[
            ::-1].reshape((1, 1, -1, 1))
        self.h1_col = cp.array(h1_col).astype('float32')[
            ::-1].reshape((1, 1, -1, 1))
        self.h0_row = cp.array(h0_row).astype('float32')[
            ::-1].reshape((1, 1, 1, -1))
        self.h1_row = cp.array(h1_row).astype('float32')[
            ::-1].reshape((1, 1, 1, -1))

    def apply(self, x):
        """ Forward pass of the DWT.

        Args:
            x (array): Input of shape :math:`(N, C_{in}, H_{in}, W_{in})`

        Returns:
            (yl, yh)
                tuple of lowpass (yl) and bandpass (yh) coefficients.
                yh is a list of scale coefficients. yl has shape
                :math:`(N, C_{in}, H_{in}', W_{in}')` and yh has shape
                :math:`list(N, C_{in}, 3, H_{in}'', W_{in}'')`. The new
                dimension in yh iterates over the LH, HL and HH coefficients.

        Note:
            :math:`H_{in}', W_{in}', H_{in}'', W_{in}''` denote the correctly
            downsampled shapes of the DWT pyramid.
        """
        # Do a multilevel transform
        # Do 1 level of the transform
        lohi = afb1d(x, self.h0_row, self.h1_row, dim=3)
        y = afb1d(lohi, self.h0_col, self.h1_col, dim=2)
        s = y.shape
        y = y.reshape(s[0], -1, 4, s[-2], s[-1])  # pylint: disable=E1121 # this might blow up in the future
        x = cp.ascontiguousarray(y[:, :, 0])
        yh = cp.ascontiguousarray(y[:, :, 1:])
        return x, yh


class DWTInverse():
    """ Performs a 2d DWT Inverse reconstruction of an image

    Args:
        wave (str): Which wavelet to use.            
    """

    def __init__(self, wave='db1'):
        super().__init__()
        wave = pywt.Wavelet(wave)
        g0_col, g1_col = wave.rec_lo, wave.rec_hi
        g0_row, g1_row = g0_col, g1_col
        # Prepare the filters
        self.g0_col = cp.array(g0_col).astype('float32').reshape((1, 1, -1, 1))
        self.g1_col = cp.array(g1_col).astype('float32').reshape((1, 1, -1, 1))
        self.g0_row = cp.array(g0_row).astype('float32').reshape((1, 1, 1, -1))
        self.g1_row = cp.array(g1_row).astype('float32').reshape((1, 1, 1, -1))

    def apply(self, coeffs):
        """
        Args:
            coeffs (yl, yh): tuple of lowpass and bandpass coefficients, where:
              yl is a lowpass array of shape :math:`(N, C_{in}, H_{in}',
              W_{in}')` and yh is a list of bandpass arrays of shape
              :math:`list(N, C_{in}, 3, H_{in}'', W_{in}'')`. I.e. should match
              the format returned by DWTForward

        Returns:
            Reconstructed input of shape :math:`(N, C_{in}, H_{in}, W_{in})`

        Note:
            :math:`H_{in}', W_{in}', H_{in}'', W_{in}''` denote the correctly
            downsampled shapes of the DWT pyramid.

        """
        yl, yh = coeffs
        lh = yh[:, :, 0]
        hl = yh[:, :, 1]
        hh = yh[:, :, 2]
        lo = sfb1d(yl, lh, self.g0_col, self.g1_col, dim=2)
        hi = sfb1d(hl, hh, self.g0_col, self.g1_col, dim=2)
        yl = sfb1d(lo, hi, self.g0_row, self.g1_row, dim=3)
        return yl


def remove_stripe_fw(data, sigma, wname, level):
    """Remove stripes with wavelet filtering"""

    [nproj, nz, ni] = data.shape

    nproj_pad = nproj + nproj // 8
    xshift = int((nproj_pad - nproj) // 2)

    # Accepts all wave types available to PyWavelets
    xfm = DWTForward(wave=wname)
    ifm = DWTInverse(wave=wname)

    # Wavelet decomposition.
    cc = []
    sli = cp.zeros([nz, 1, nproj_pad, ni], dtype='float32')

    sli[:, 0, (nproj_pad - nproj)//2:(nproj_pad + nproj) //
        2] = data.astype('float32').swapaxes(0, 1)
    for k in range(level):
        sli, c = xfm.apply(sli)
        cc.append(c)
        # FFT
        fcV = cp.fft.fft(cc[k][:, 0, 1], axis=1)
        _, my, mx = fcV.shape
        # Damping of ring artifact information.
        y_hat = cp.fft.ifftshift((cp.arange(-my, my, 2) + 1) / 2)
        damp = -cp.expm1(-y_hat**2 / (2 * sigma**2))
        fcV *= cp.tile(damp, (mx, 1)).swapaxes(0, 1)
        # Inverse FFT.
        cc[k][:, 0, 1] = cp.fft.ifft(fcV, my, axis=1).real

    # Wavelet reconstruction.
    for k in range(level)[::-1]:
        shape0 = cc[k][0, 0, 1].shape
        sli = sli[:, :, :shape0[0], :shape0[1]]
        sli = ifm.apply((sli, cc[k]))

    data = sli[:, 0, (nproj_pad - nproj)//2:(nproj_pad + nproj) //
               2, :ni].astype(data.dtype)  # modified
    data = data.swapaxes(0, 1)

    return data

######## Titarenko ring removal ############################################################################################################################################################################
def remove_stripe_ti(data, beta, mask_size):
    """Remove stripes with a new method by V. Titareno """
    gamma = beta*((1-beta)/(1+beta)
                  )**cp.abs(cp.fft.fftfreq(data.shape[-1])*data.shape[-1])
    gamma[0] -= 1
    v = cp.mean(data, axis=0)
    v = v-v[:, 0:1]
    v = cp.fft.irfft(cp.fft.rfft(v)*cp.fft.rfft(gamma))
    mask = cp.zeros(v.shape, dtype=v.dtype)
    mask_size = mask_size*mask.shape[1]
    mask[:, mask.shape[1]//2-mask_size//2:mask.shape[1]//2+mask_size//2] = 1
    data[:] += v*mask
    return data


######## Optimized version for Vo-all ring removal in tomopy################################################################################################################################################################
def _rs_sort(sinogram, size, matindex, dim):
    """
    Remove stripes using the sorting technique.
    """
    sinogram = cp.transpose(sinogram)
    matcomb = cp.asarray(cp.dstack((matindex, sinogram)))
    
    # matsort = cp.asarray([row[row[:, 1].argsort()] for row in matcomb])
    ids = cp.argsort(matcomb[:,:,1],axis=1)
    matsort = matcomb.copy()
    matsort[:,:,0] = cp.take_along_axis(matsort[:,:,0],ids,axis=1)
    matsort[:,:,1] = cp.take_along_axis(matsort[:,:,1],ids,axis=1)
    if dim == 1:
        matsort[:, :, 1] = median_filter(matsort[:, :, 1], (size, 1))
    else:
        matsort[:, :, 1] = median_filter(matsort[:, :, 1], (size, size))
    
    # matsortback = cp.asarray([row[row[:, 0].argsort()] for row in matsort])
    
    ids = cp.argsort(matsort[:,:,0],axis=1)
    matsortback = matsort.copy()
    matsortback[:,:,0] = cp.take_along_axis(matsortback[:,:,0],ids,axis=1)
    matsortback[:,:,1] = cp.take_along_axis(matsortback[:,:,1],ids,axis=1)
    
    sino_corrected = matsortback[:, :, 1]
    return cp.transpose(sino_corrected)

def _mpolyfit(x,y):
    n= len(x)
    x_mean = cp.mean(x)
    y_mean = cp.mean(y)
    
    Sxy = cp.sum(x*y) - n*x_mean*y_mean
    Sxx = cp.sum(x*x) - n*x_mean*x_mean
    
    slope = Sxy / Sxx
    intercept = y_mean - slope*x_mean
    return slope,intercept

def _detect_stripe(listdata, snr):
    """
    Algorithm 4 in :cite:`Vo:18`. Used to locate stripes.
    """
    numdata = len(listdata)
    listsorted = cp.sort(listdata)[::-1]
    xlist = cp.arange(0, numdata, 1.0)
    ndrop = cp.int16(0.25 * numdata)
    # (_slope, _intercept) = cp.polyfit(xlist[ndrop:-ndrop - 1],
                                    #   listsorted[ndrop:-ndrop - 1], 1)
    (_slope, _intercept) = _mpolyfit(xlist[ndrop:-ndrop - 1], listsorted[ndrop:-ndrop - 1])

    numt1 = _intercept + _slope * xlist[-1]
    noiselevel = cp.abs(numt1 - _intercept)
    noiselevel = cp.clip(noiselevel, 1e-6, None)
    val1 = cp.abs(listsorted[0] - _intercept) / noiselevel
    val2 = cp.abs(listsorted[-1] - numt1) / noiselevel
    listmask = cp.zeros_like(listdata)
    if (val1 >= snr):
        upper_thresh = _intercept + noiselevel * snr * 0.5
        listmask[listdata > upper_thresh] = 1.0
    if (val2 >= snr):
        lower_thresh = numt1 - noiselevel * snr * 0.5
        listmask[listdata <= lower_thresh] = 1.0
    return listmask

def _rs_large(sinogram, snr, size, matindex, drop_ratio=0.1, norm=True):
    """
    Remove large stripes.
    """
    drop_ratio = max(min(drop_ratio,0.8),0)# = cp.clip(drop_ratio, 0.0, 0.8)
    (nrow, ncol) = sinogram.shape
    ndrop = int(0.5 * drop_ratio * nrow)
    sinosort = cp.sort(sinogram, axis=0)
    sinosmooth = median_filter(sinosort, (1, size))
    list1 = cp.mean(sinosort[ndrop:nrow - ndrop], axis=0)
    list2 = cp.mean(sinosmooth[ndrop:nrow - ndrop], axis=0)
    # listfact = cp.divide(list1,
    #                      list2,
    #                      out=cp.ones_like(list1),
    #                      where=list2 != 0)
    
    listfact = list1/list2
    
    # Locate stripes
    listmask = _detect_stripe(listfact, snr)
    listmask = binary_dilation(listmask, iterations=1).astype(listmask.dtype)
    matfact = cp.tile(listfact, (nrow, 1))
    # Normalize
    if norm is True:
        sinogram = sinogram / matfact
    sinogram1 = cp.transpose(sinogram)
    matcombine = cp.asarray(cp.dstack((matindex, sinogram1)))
    
    # matsort = cp.asarray([row[row[:, 1].argsort()] for row in matcombine])
    ids = cp.argsort(matcombine[:,:,1],axis=1)
    matsort = matcombine.copy()
    matsort[:,:,0] = cp.take_along_axis(matsort[:,:,0],ids,axis=1)
    matsort[:,:,1] = cp.take_along_axis(matsort[:,:,1],ids,axis=1)
    
    matsort[:, :, 1] = cp.transpose(sinosmooth)
    # matsortback = cp.asarray([row[row[:, 0].argsort()] for row in matsort])
    ids = cp.argsort(matsort[:,:,0],axis=1)
    matsortback = matsort.copy()
    matsortback[:,:,0] = cp.take_along_axis(matsortback[:,:,0],ids,axis=1)
    matsortback[:,:,1] = cp.take_along_axis(matsortback[:,:,1],ids,axis=1)
    
    sino_corrected = cp.transpose(matsortback[:, :, 1])
    listxmiss = cp.where(listmask > 0.0)[0]
    sinogram[:, listxmiss] = sino_corrected[:, listxmiss]
    return sinogram

def _rs_dead(sinogram, snr, size, matindex, norm=True):
    """
    Remove unresponsive and fluctuating stripes.
    """
    sinogram = cp.copy(sinogram)  # Make it mutable
    (nrow, _) = sinogram.shape
    # sinosmooth = cp.apply_along_axis(uniform_filter1d, 0, sinogram, 10)
    sinosmooth = uniform_filter1d(sinogram, 10, axis=0)
    
    listdiff = cp.sum(cp.abs(sinogram - sinosmooth), axis=0)
    listdiffbck = median_filter(listdiff, size)
    
    
    listfact = listdiff/listdiffbck
    
    listmask = _detect_stripe(listfact, snr)
    listmask = binary_dilation(listmask, iterations=1).astype(listmask.dtype)
    listmask[0:2] = 0.0
    listmask[-2:] = 0.0
    listx = cp.where(listmask < 1.0)[0]
    listy = cp.arange(nrow)
    matz = sinogram[:, listx]
    
    listxmiss = cp.where(listmask > 0.0)[0]    
    
    # finter = interpolate.interp2d(listx.get(), listy.get(), matz.get(), kind='linear')    
    if len(listxmiss) > 0:
        # sinogram_c[:, listxmiss.get()] = finter(listxmiss.get(), listy.get())        
        ids = cp.searchsorted(listx, listxmiss)
        sinogram[:,listxmiss] = matz[:,ids-1]+(listxmiss-listx[ids-1])*(matz[:,ids]-matz[:,ids-1])/(listx[ids]-listx[ids-1])
        
    # Remove residual stripes
    if norm is True:
        sinogram = _rs_large(sinogram, snr, size, matindex)
    return sinogram


def _create_matindex(nrow, ncol):
    """
    Create a 2D array of indexes used for the sorting technique.
    """
    listindex = cp.arange(0.0, ncol, 1.0)
    matindex = cp.tile(listindex, (nrow, 1))
    return matindex

def remove_all_stripe(tomo,
                      snr=3,
                      la_size=61,
                      sm_size=21,
                      dim=1):
    """
    Remove all types of stripe artifacts from sinogram using Nghia Vo's
    approach :cite:`Vo:18` (combination of algorithm 3,4,5, and 6).

    Parameters
    ----------
    tomo : ndarray
        3D tomographic data.
    snr  : float
        Ratio used to locate large stripes.
        Greater is less sensitive.
    la_size : int
        Window size of the median filter to remove large stripes.
    sm_size : int
        Window size of the median filter to remove small-to-medium stripes.
    dim : {1, 2}, optional
        Dimension of the window.

    Returns
    -------
    ndarray
        Corrected 3D tomographic data.
    """
    matindex = _create_matindex(tomo.shape[2], tomo.shape[0])
    for m in range(tomo.shape[1]):
        sino = tomo[:, m, :]
        sino = _rs_dead(sino, snr, la_size, matindex)
        sino = _rs_sort(sino, sm_size, matindex, dim)
        tomo[:, m, :] = sino
    return tomo


def remove_all_stripe_sinos(sinos, snr=3, la_size=61, sm_size=21, dim=1):
    """
    Same as remove_all_stripe(), but acting on sinograms
    """
    n_sinos, n_a, n_x = sinos.shape
    matindex = _create_matindex(n_x, n_a)
    for m in range(n_sinos):
        sino = sinos[m]
        sino = _rs_dead(sino, snr, la_size, matindex)
        sino = _rs_sort(sino, sm_size, matindex, dim)
        sinos[m] = sino
    return sinos


from ..cuda.utils import pycuda_to_cupy
def remove_all_stripe_pycuda(array, layout="radios", device_id=0, **kwargs):
    """
    Nabu interface to tomocupy "remove_all_stripe".
    Processing is done in-place to save memory, meaning that the content of "array" will be overwritten.

    Parameters
    ----------
    array: pycuda.GPUArray
        Stack of radios in the shape (n_angles, n_y, n_x), if layout == "radios"
        Stack of sinos in the shape (n_y, n_angles, n_x), if layout == "sinos".


    Other Parameters
    ----------------
    See parameters of 'remove_all_stripe
    """

    # Init cupy. Nabu currently does not use cupy, with exception of this module,
    # so the initialization has to be done here.
    if getattr(remove_all_stripe, "_cupy_init", False) is False:
        from cupy import cuda
        cuda.Device(device_id).use()
        setattr(remove_all_stripe, "_cupy_init", True)

    # remove_all_stripe() in tomocupy expects a 3D array to build the "matindex" data structure.
    # The usage of this "matindex" array is not clear since this method is supposed to act on individual sinograms.
    # To avoid memory duplication, we use fake 3D array, i.e, we pass a series of (1, n_a, n_x) sinograms

    if layout == "radios":
        sinos = array.transpose(axes=(1, 0, 2)) # no copy
    else:
        sinos = array
    # is_contiguous = sinos.flags.c_contigious
    n_sinos, n_a, n_x = sinos.shape
    sinos_tmp = garray.zeros((1, n_a, n_x), dtype="f")
    for i in range(n_sinos):
        sinos_tmp[0] = sinos[i]
        cupy_sinos = pycuda_to_cupy(sinos_tmp) # no memory copy, the internal pointer is passed to pycuda
        remove_all_stripe_sinos(cupy_sinos, **kwargs)
        sinos[i] = sinos_tmp[0]
    return array



from ..processing.rotation import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.rotation has been moved to nabu.processing.rotation", do_print=True, func_name="rotation"
)
